import os
import sys
import glob
import numpy as np
import pandas as pd
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import matplotlib

matplotlib.use('agg')
import tensorflow as tf
import tensorflow.keras as K
import scipy
import tables
tf.keras.backend.set_epsilon(0.0000001)
tf_version = tf.__version__  # ToDo use packaging.version

if tf_version <= '1.13.1':
    from gennet.utils.LocallyDirected1D import LocallyDirected1D
elif tf_version >= '2.0':
    from gennet.utils.LocallyDirected1D import LocallyDirected1D
else:
    print("unexpected tensorflow version")
    from gennet.utils.LocallyDirected1D import LocallyDirected1D
    
    
def example_network():
    mask = scipy.sparse.load_npz('./folder/snps_gene.npz')
    masks = [mask]
    
    inputs_ = K.Input((mask.shape[0],), name='inputs_')
    input_cov = K.Input((num_covariates,), name='inputs_cov')
    
    layer_0 = K.layers.Reshape(input_shape=(mask.shape[0],), target_shape=(inputsize, 1))(inputs_)
    
    layer_1 = LocallyDirected1D(mask=mask, filters=1, input_shape=(inputsize, 1), name="gene_layer")(layer_0)
    layer_1 = K.layers.Flatten()(layer_1)
    layer_1 = K.layers.Activation("relu")(layer_1)
    layer_1 = K.layers.BatchNormalization(center=False, scale=False, name="inter_out")(layer_1)
    
    layer_2 = K.layers.Dense(units=1)(layer_1)
    layer_2 = K.layers.Activation("relu")(layer_2)
    model = K.Model(inputs=[inputs_, input_cov], outputs=layer_2)
    print(model.summary())
    
    return model, masks


def regression_properties(datapath):
    ytrain = pd.read_csv(datapath + "subjects.csv")
    mean_ytrain = float(ytrain[ytrain["set"] == 1]["labels"].mean())
    negative_values_ytrain = float(ytrain[ytrain["set"] == 1]["labels"].min()) < 0
    return mean_ytrain, negative_values_ytrain


def layer_block(model, mask, i, regression, L1_act =0.01):
    if regression:
        activation_type="relu"
    else:
        activation_type="tanh"
    
    model = LocallyDirected1D(mask=mask, filters=1, input_shape=(mask.shape[0], 1),
                              name="LocallyDirected_" + str(i), activity_regularizer=K.regularizers.l1(L1_act))(model)
    model = K.layers.Activation(activation_type)(model)
    model = K.layers.BatchNormalization(center=False, scale=False)(model)
    return model


def activation_layer(model, regression, negative_values_ytrain):
   
    if regression: 
        if negative_values_ytrain:
            model = K.layers.Activation("linear")(model)
            print('using a linear activation function')
        else:
            model = K.layers.Activation("relu")(model)
            print('using a relu activation function')
    else:
        model = K.layers.Activation("sigmoid")(model)
        
    return model

def one_hot_input(input_layer):
    model = K.layers.LocallyConnected1D(filters=1, strides=1, kernel_size=1, 
                                            name="SNP_layer", implementation=3
                                           )(input_layer)
    model = K.layers.Activation('linear', name="snp_activation")(model) 
    return model
    

def add_covariates(model, input_cov, num_covariates, regression, negative_values_ytrain, mean_ytrain, l1_value, L1_act):
    if num_covariates > 0:
        model = activation_layer(model, regression, negative_values_ytrain)
        model = K.layers.concatenate([model, input_cov], axis=1, name="concatenate_cov")
        model = K.layers.BatchNormalization(center=False, scale=False, name="batchnorm_cov")(model)
        model = K.layers.Dense(units=1, name="output_layer_cov",
                       kernel_regularizer=tf.keras.regularizers.l1(l=l1_value),
                       activity_regularizer=K.regularizers.l1(L1_act),
                       bias_initializer= tf.keras.initializers.Constant(mean_ytrain))(model)
    return model

def create_network_from_npz(datapath,
                            inputsize,
                            genotype_path,
                            l1_value=0.01,
                            L1_act =0.01,
                            regression=False,
                            one_hot = False,
                            num_covariates=0,
                            mask_order = []):
    print("Creating networks from npz masks")
    print("regression", regression)
    print("one_hot", one_hot)
    if regression:
        mean_ytrain, negative_values_ytrain = regression_properties(datapath)
    else:
        mean_ytrain = 0
        negative_values_ytrain = False

    masks = []
    mask_shapes_x = []
    mask_shapes_y = []

    print(mask_order)

    if len(mask_order) > 0:  # if mask_order is defined we use this order
        for mask in mask_order:
            mask = scipy.sparse.load_npz(datapath + '/'+str(mask)+'.npz')
            masks.append(mask)
            mask_shapes_x.append(mask.shape[0])
            mask_shapes_y.append(mask.shape[1])

        for x in range(len(masks) - 1):  # check that the masks fit eachother
            assert mask_shapes_y[x] == mask_shapes_x[x + 1]
    else:
        # if mask order is not defined we can sort the mask by the size
        for npz_path in glob.glob(datapath + '/*.npz'):
            mask = scipy.sparse.load_npz(npz_path)
            masks.append(mask)
            mask_shapes_x.append(mask.shape[0])
            mask_shapes_y.append(mask.shape[1])

        for i in range(len(masks)):  # sort all the masks in the correct order
            argsort_x = np.argsort(mask_shapes_x)[::-1]
            argsort_y = np.argsort(mask_shapes_y)[::-1]

            mask_shapes_x = np.array(mask_shapes_x)
            mask_shapes_y = np.array(mask_shapes_y)
            assert all(argsort_x == argsort_y)  # check that both dimensions have the same order

            masks = [masks[i] for i in argsort_y]  # sort masks
            mask_shapes_x = mask_shapes_x[argsort_x]
            mask_shapes_y = mask_shapes_y[argsort_y]

            for x in range(len(masks) - 1):  # check that the masks fit eachother
                assert mask_shapes_y[x] == mask_shapes_x[x + 1]

    assert mask_shapes_x[0] == inputsize
    if mask_shapes_y[-1] == 1:  # should we end with a dense layer?
        all_masks_available = True
    else:
        all_masks_available = False

    input_cov = K.Input((num_covariates,), name='inputs_cov')
    
    if one_hot:
        input_layer = K.Input((inputsize, 3), name='input_layer')
        model = one_hot_input(input_layer)
    else:
        input_layer = K.Input((inputsize,), name='input_layer')
        model = K.layers.Reshape(input_shape=(inputsize,), target_shape=(inputsize, 1))(input_layer)

    for i in range(len(masks)):
        mask = masks[i]
        model = layer_block(model, mask, i, regression, L1_act=L1_act)

    model = K.layers.Flatten()(model)

    if all_masks_available:
        model = LocallyDirected1D(mask=masks[-1], filters=1, input_shape=(mask.shape[0], 1),
                                  name="output_layer")(model)
    else:
        model = K.layers.Dense(units=1, name="output_layer",
                               kernel_regularizer=tf.keras.regularizers.l1(l=l1_value),
                               activity_regularizer=K.regularizers.l1(L1_act),
                               bias_initializer= tf.keras.initializers.Constant(mean_ytrain))(model)

    model = add_covariates(model, input_cov, num_covariates, regression, negative_values_ytrain, mean_ytrain, l1_value, L1_act)

    output_layer = activation_layer(model, regression, negative_values_ytrain)
    model = K.Model(inputs=[input_layer, input_cov], outputs=output_layer)

    print(model.summary())

    return model, masks



def create_network_from_csv(datapath, 
                            inputsize, 
                            genotype_path,
                            l1_value=0.01, 
                            L1_act =0.01, 
                            regression=False,
                            one_hot=False,
                            num_covariates=0):
    
    print("Creating networks from npz masks")
    print("regression", regression)
    if regression:
        mean_ytrain, negative_values_ytrain = regression_properties(datapath)
        print('mean_ytrain',mean_ytrain)
        print('negative_values_ytrain',negative_values_ytrain)
    else:
        mean_ytrain = 0
        negative_values_ytrain = False
        
        
    masks = []
    
    network_csv = pd.read_csv(datapath + "/topology.csv")
    network_csv = network_csv.filter(like="node", axis=1)
    columns = list(network_csv.columns.values)
    network_csv = network_csv.sort_values(by=columns, ascending=True)

    input_cov = K.Input((num_covariates,), name='inputs_cov')
    
    if one_hot:
        input_layer = K.Input((inputsize, 3), name='input_layer')
        model = one_hot_input(input_layer)
    else:
        input_layer = K.Input((inputsize,), name='input_layer')
        model = K.layers.Reshape(input_shape=(inputsize,), target_shape=(inputsize, 1))(input_layer)

    for i in range(len(columns) - 1):
        matrix_ones = np.ones(len(network_csv[[columns[i], columns[i + 1]]]), bool)
        matrix_coord = (network_csv[columns[i]].values, network_csv[columns[i + 1]].values)
        if i == 0:
            matrixshape = (inputsize, network_csv[columns[i + 1]].max() + 1)
        else:
            matrixshape = (network_csv[columns[i]].max() + 1, network_csv[columns[i + 1]].max() + 1)
        mask = scipy.sparse.coo_matrix(((matrix_ones), matrix_coord), shape = matrixshape)
        masks.append(mask)
        model = layer_block(model, mask, i, regression, L1_act=L1_act)

    model = K.layers.Flatten()(model)

    model = K.layers.Dense(units=1, name="output_layer",
                           kernel_regularizer=tf.keras.regularizers.l1(l=l1_value),
                           activity_regularizer=K.regularizers.l1(L1_act),
                           bias_initializer= tf.keras.initializers.Constant(mean_ytrain))(model)
    
    model = add_covariates(model, input_cov, num_covariates, regression, negative_values_ytrain, mean_ytrain, l1_value, L1_act)
    
    output_layer = activation_layer(model, regression, negative_values_ytrain)
   

    model = K.Model(inputs=[input_layer, input_cov], outputs=output_layer)

    print(model.summary())

    return model, masks


def lasso(inputsize, l1_value, num_covariates=0, regression=False):
    masks=[]
    inputs = K.Input((inputsize,), name='inputs')
    input_cov = K.Input((num_covariates,), name='inputs_cov')
    model = K.layers.BatchNormalization(center=False, scale=False, name="inter_out")(inputs)
    model = K.layers.Dense(units=1, kernel_regularizer=K.regularizers.l1(l1_value))(model)
    
    model = add_covariates(model, input_cov, num_covariates, regression, negative_values_ytrain, mean_ytrain, l1_value, L1_act)
    
    output_layer = K.layers.Activation("sigmoid")(model)
    
    model = K.Model(inputs=[inputs, input_cov], outputs=output_layer)
    return model, masks


def sparse_directed_gene_l1(datapath, inputsize, l1_value=0.01, L1_act=0.01, one_hot=False):
    
    
    for npz_path in glob.glob(datapath + '/*.npz'):
        mask = scipy.sparse.load_npz(npz_path)
    masks = [mask]     
    
    
    if one_hot:
        input_layer = K.Input((inputsize, 3), name='input_layer')
        model = one_hot_input(input_layer)
    else:
        input_layer = K.Input((inputsize,), name='input_layer')
        model = K.layers.Reshape(input_shape=(inputsize,), target_shape=(inputsize, 1))(input_layer)
    


    x1_1 = LocallyDirected1D(mask=mask, filters=1, input_shape=(inputsize, 1)
                                    , name="gene_layer", activity_regularizer=K.regularizers.l1(L1_act) )(model)
    x1_out = K.layers.Flatten()(x1_1)
    x1_out = K.layers.Activation("tanh")(x1_out)
    x1_out = K.layers.BatchNormalization(center=False, scale=False, name="inter_out")(x1_out)

    x9 = K.layers.Dense(units=1,kernel_regularizer=K.regularizers.l1(l1_value),  
                        activity_regularizer=K.regularizers.l1(L1_act))(x1_out)
    x9 = K.layers.Activation("sigmoid")(x9)


    model = K.Model(inputs=input_layer, outputs=x9)
    
    print(model.summary())
    
    return model, masks


def gene_network_multiple_filters(datapath, 
                                  inputsize,
                                  genotype_path, 
                                  l1_value=0.01,
                                  L1_act =0.01,
                                  regression=False,
                                  num_covariates=0, 
                                  filters=2,
                                  one_hot=False):
    
    print("Creating networks from npz masks")
    print("regression", regression)
    if regression:
        mean_ytrain, negative_values_ytrain = regression_properties(datapath)
        print('mean_ytrain',mean_ytrain)
        print('negative_values_ytrain',negative_values_ytrain)
    else:
        mean_ytrain = 0
        negative_values_ytrain = False
        
    print("gene_network_multiple_filters with", filters, "filters")
    
    masks = []
    for npz_path in glob.glob(datapath + '/*.npz'):
        mask = scipy.sparse.load_npz(npz_path)
        masks.append(mask)
    if len(masks) == 0:
        print("You need an npz mask to run this network. Convert topology.csv to a mask.npz") 
        exit()
    if len(masks) > 1:
        print("multiple masks found")
    
    input_cov = K.Input((num_covariates,), name='inputs_cov')
    
    if one_hot:
        input_layer = K.Input((inputsize, 3), name='input_layer')
        model = one_hot_input(input_layer)
    else:
        input_layer = K.Input((inputsize,), name='input_layer')
        model = K.layers.Reshape(input_shape=(inputsize,), target_shape=(inputsize, 1))(input_layer)
    
    model = LocallyDirected1D(mask=mask, filters=filters, input_shape=(inputsize, 1), name="gene_layer")(model)
    model = K.layers.Activation("relu")(model)
    model = K.layers.BatchNormalization(center=False, scale=False, name="inter_out_g1")(model)
    
    model = K.layers.LocallyConnected1D(filters=1, strides=1, kernel_size=1, implementation=3)(model)
    model = K.layers.Activation("relu")(model)
    model = K.layers.BatchNormalization(center=False, scale=False, name="inter_out_g2")(model)
    
    model = K.layers.Flatten()(model)
    model = K.layers.Dense(units=1, name="output_layer",
                           kernel_regularizer=tf.keras.regularizers.l1(l=l1_value), 
                           bias_initializer= tf.keras.initializers.Constant(mean_ytrain))(model)
    
    model = add_covariates(model, input_cov, num_covariates, regression, 
                           negative_values_ytrain, mean_ytrain, l1_value, L1_act)
    
    output_layer = activation_layer(model, regression, negative_values_ytrain)
   

    model = K.Model(inputs=[input_layer, input_cov], outputs=output_layer)

    print(model.summary())
    return model, masks


def gene_network_snp_gene_filters(datapath, 
                                  inputsize,
                                  genotype_path, 
                                  l1_value=0.01, 
                                  L1_act =0.01,
                                  regression=False,
                                  num_covariates=0, 
                                  filters=2,
                                  one_hot=False):
    
    print("Creating networks from npz masks")
    print("regression", regression)
    if regression:
        mean_ytrain, negative_values_ytrain = regression_properties(datapath)
        print('mean_ytrain',mean_ytrain)
        print('negative_values_ytrain',negative_values_ytrain)
    else:
        mean_ytrain = 0
        negative_values_ytrain = False
        
    print("height_multiple_filters with", filters, "filters")
    
    masks = []
    for npz_path in glob.glob(datapath + '/*.npz'):
        mask = scipy.sparse.load_npz(npz_path)
        masks.append(mask)
    if len(masks) == 0:
        print("You need an npz mask to run this network. Convert topology.csv to a mask.npz")
        exit()
    
    if len(masks) > 1:
        print("multiple masks found")
    
    input_cov = K.Input((num_covariates,), name='inputs_cov')
    
    if one_hot:
        input_layer = K.Input((inputsize, 3), name='input_layer')
        model = one_hot_input(input_layer)
    else:
        input_layer = K.Input((inputsize,), name='input_layer')
        model = K.layers.Reshape(input_shape=(inputsize,), target_shape=(inputsize, 1))(input_layer)
    
    
    model = LocallyDirected1D(mask=mask, filters=filters, input_shape=(inputsize, 1), name="gene_layer")(model)
    model = K.layers.Activation("relu")(model)
    model = K.layers.BatchNormalization(center=False, scale=False, name="inter_out_g1")(model)
    
    model = K.layers.LocallyConnected1D(filters=filters, strides=1, kernel_size=1, implementation=3)(model)
    model = K.layers.Activation("relu")(model)
    model = K.layers.BatchNormalization(center=False, scale=False, name="inter_out_g2")(model)
    
    model = K.layers.LocallyConnected1D(filters=1, strides=1, kernel_size=1, implementation=3)(model)
    model = K.layers.Activation("relu")(model)
    model = K.layers.BatchNormalization(center=False, scale=False, name="inter_out_g3")(model)
    
    model = K.layers.Flatten()(model)
    model = K.layers.Dense(units=1, name="output_layer",
                           kernel_regularizer=tf.keras.regularizers.l1(l=l1_value), 
                           bias_initializer= tf.keras.initializers.Constant(mean_ytrain))(model)
    
    model = add_covariates(model, input_cov, num_covariates, regression, negative_values_ytrain, mean_ytrain, l1_value, L1_act)
    
    output_layer = activation_layer(model, regression, negative_values_ytrain)
   

    model = K.Model(inputs=[input_layer, input_cov], outputs=output_layer)

    print(model.summary())
    return model, masks



def regression_height(inputsize, num_covariates=2, l1_value=0.001):
    mask = scipy.sparse.load_npz('/home/ahilten/repositories/pheno_height/Input_files/SNP_nearest_gene_mask.npz')
    masks = [mask]
    
    input_cov = K.Input((num_covariates,), name='inputs_cov')
    
    inputs_ = K.Input((mask.shape[0],), name='inputs_')
    layer_0 = K.layers.Reshape(input_shape=(mask.shape[0],), target_shape=(inputsize, 1))(inputs_)
    
    layer_1 = LocallyDirected1D(mask=mask, filters=1, input_shape=(inputsize, 1), name="gene_layer")(layer_0)
    layer_1 = K.layers.Flatten()(layer_1)
    layer_1 = K.layers.Activation("relu")(layer_1)
    layer_1 = K.layers.BatchNormalization()(layer_1)
    
    layer_2 = K.layers.Dense(units=10, kernel_regularizer=tf.keras.regularizers.l1(l=l1_value))(layer_1)
    layer_2 = K.layers.Activation("relu")(layer_2) 
    
    layer_3 = K.layers.concatenate([layer_2, input_cov], axis=1)
    layer_3 = K.layers.BatchNormalization()(layer_3)
    layer_3 = K.layers.Dense(units=10)(layer_3)
    layer_3 = K.layers.Activation("relu")(layer_3) 
        
    layer_4 = K.layers.Dense(units=10)(layer_3)
    layer_4 = K.layers.Activation("relu")(layer_4) 
    
    layer_5 = K.layers.Dense(units=1, bias_initializer= tf.keras.initializers.Constant(168))(layer_4)
    layer_5 = K.layers.Activation("relu")(layer_5)
    
    model = K.Model(inputs=[inputs_, input_cov], outputs=layer_5)
    
    print(model.summary())
    
    return model, masks
    


def remove_batchnorm_model(model, masks, keep_cov = False):
    original_model = model
    inputs = tf.keras.Input(shape=original_model.input_shape[0][1:])
    x = inputs

    mask_num = 0
    for layer in original_model.layers[1:]: 
        # Skip BatchNormalization layers
        if not isinstance(layer, tf.keras.layers.BatchNormalization):
            # Handle LocallyDirected1D layer with custom arguments
            if isinstance(layer, LocallyDirected1D):
                config = layer.get_config()
                new_layer = LocallyDirected1D(filters=config['filters'], 
                                                mask=masks[mask_num],
                                                name=config['name'])
                x = new_layer(x)
                mask_num = mask_num + 1
            elif "_cov" in layer.name and not keep_cov:
                pass
            else:
                # Add other layers as they are
                x = layer.__class__.from_config(layer.get_config())(x)

    # Create the new model
    new_model = tf.keras.Model(inputs=inputs, outputs=x)

    original_model_layers = [x for x in original_model.layers if not isinstance(x, tf.keras.layers.BatchNormalization)]

    for new_layer, layer in zip(new_model.layers, original_model_layers): 
        new_layer.set_weights(layer.get_weights())

    print(new_model.summary())
        
    return new_model


def remove_cov(model, masks):
    original_model = model
    inputs = tf.keras.Input(shape=original_model.input_shape[0][1:])
    x = inputs

    mask_num = 0
    for layer in original_model.layers[1:]: 
        # Skip BatchNormalization layers
        if isinstance(layer, LocallyDirected1D):
            config = layer.get_config()
            new_layer = LocallyDirected1D(filters=config['filters'], 
                                            mask=masks[mask_num],
                                            name=config['name'])
            x = new_layer(x)
            mask_num = mask_num + 1
        elif "_cov" in layer.name:
            pass
        else:
            # Add other layers as they are
            x = layer.__class__.from_config(layer.get_config())(x)

    # Create the new model
    new_model = tf.keras.Model(inputs=inputs, outputs=x)

    for new_layer, layer in zip(new_model.layers, original_model.layers ): 
        new_layer.set_weights(layer.get_weights())

    print(new_model.summary())
        
    return new_model



import os

_proc_status = '/proc/%d/status' % os.getpid()

_scale = {'kB': 1024.0, 'mB': 1024.0 * 1024.0,
          'KB': 1024.0, 'MB': 1024.0 * 1024.0}


def _VmB(VmKey):
    '''Private.
    '''
    global _proc_status, _scale
    # get pseudo file  /proc/<pid>/status
    try:
        t = open(_proc_status)
        v = t.read()
        t.close()
    except:
        return 0.0  # non-Linux?
    # get VmKey line e.g. 'VmRSS:  9999  kB\n ...'
    i = v.index(VmKey)
    v = v[i:].split(None, 3)  # whitespace
    if len(v) < 3:
        return 0.0  # invalid format?
    # convert Vm value to bytes
    return float(v[1]) * _scale[v[2]]


def memory(since=0.0):
    '''Return memory usage in bytes.
    '''
    return _VmB('VmSize:') - since


def resident(since=0.0):
    '''Return resident memory usage in bytes.
    '''
    return _VmB('VmRSS:') - since



import argparse

parser = argparse.ArgumentParser(description='Script to merge genotype')
parser.add_argument("-g", "--genotype", nargs='+', type=str, help="path/paths to genotype data folder")
parser.add_argument('-mapper', type=str, help='Mapper data folder')
parser.add_argument('-mapper_name', type=str, help='Mapper name')
parser.add_argument("-o", "--out", type=str, required=True, help="path to save result folder")
parser.add_argument("-save_name", type=str, required=True, help="merge study name")
parser.add_argument('-study_name', type=str, required=True, nargs='+',
                    help=' Name for saved genotype data, without ext')

parser.add_argument('-cluster', type=str, default='n', choices=['y', 'n'],
                    help=' Is it parallel cluster job, default no')
parser.add_argument('-node', nargs='+', help='number of nodes / this node number, example: 10 2 ')
parser.add_argument('-split', type=int, help='Split size for merge genotypes')

args = parser.parse_args()
print(args)

if __name__ == '__main__':
    print('Not implemented!')

# mapper=Mapper(args.mapper_name)
# mapper.load(args.mapper)
# mapper.chunk_size=args.split
#
#
# hdf5_iter=0
# h5_name=args.save_name
# pytable_filter=tables.Filters(complevel=9, complib='zlib')
# gen=[]
# for i,j in enumerate(args.genotype):
# 	gen.append(Reader('genotype'))
# 	gen[i].start(j,hdf5=True, study_name=args.study_name[i], ID=False)
#
# RSID=[]
# SUB_ID=[]
# for i in gen:
# 	SUB_ID.append(i.folder._data.get_id())
# mapper.cluster=args.cluster
# mapper.node=args.node
#
# while True:
# 	if args.cluster=='n':
# 		SNPs_index, keys=mapper.get_next()
# 	else:
# 		chunk=mapper.chunk_pop()
# 		if chunk is None:
# 			SNPs_index=None
# 			break
# 		print chunk
# 		SNPs_index, keys=mapper.get_chunk(chunk)
#
# 	if SNPs_index is None:
# 		break
# 	RSID.append(keys)
#
# 	data=merge_genotype(gen, SNPs_index) #TODO (high) add mapper
# 	print data.shape
# 	if args.cluster=='n':
# 		h5_gen_file = tables.open_file(
# 			os.path.join(args.out,str(hdf5_iter)+'_'+h5_name+'.h5'), 'w', title=args.save_name)
# 	else:#TODO (high) check!
# 		h5_gen_file = tables.open_file(
# 			os.path.join(args.out,str(chunk[0])+'_' +str(chunk[1])+'_'+h5_name+'.h5'), 'w', title=args.save_name)
# 	hdf5_iter+=1
#
# 	atom = tables.Int8Atom()  # TODO (low) check data format
# 	genotype = h5_gen_file.create_carray(h5_gen_file.root, 'genotype', atom,
# 										(data.shape),
# 										title='Genotype',
# 										filters=pytable_filter)
# 	genotype[:] = data
# 	h5_gen_file.close()
# 	genotype=None
# 	data=None
# 	gc.collect()
# 	print hdf5_iter
#
# RSID=np.array(RSID)
# SUB_ID=np.array(SUB_ID)
# if args.cluster=='n':
# 	np.save(os.path.join(args.out,'RSID.npy'),RSID)
# 	np.save(os.path.join(args.out,'SUB_ID.npy'),SUB_ID)
#
# else:
# 	np.save(os.path.join(args.out,str(args.node[1])+'_RSID.npy'),RSID)
# 	np.save(os.path.join(args.out,str(args.node[1])+'_SUB_ID.npy'),SUB_ID)



import os
import pytest
import pandas as pd
import shutil
from os.path import dirname, abspath
from gennet.utils.Interpret import interpret
from gennet.utils.Create_plots import  sunburst_plot, plot_layer_weight, manhattan_relative_importance
from gennet.utils.Utility_functions import get_paths

# TODO: add test without covariates
# TODO add test with covariates for regression + classification
# TODO add test with multiple genotype files.
# test randomnesss after .. epoch shuffles.
# ToDO add test for each file.

def get_GenNet_path():
    return str(dirname(dirname(abspath(__file__)))) + "/"

def remove_old(ID):
    GenNet_path  =  get_GenNet_path()
    resultpath = GenNet_path +  "results/GenNet_experiment_" + str(ID)  + '_/'
    try:
        shutil.rmtree(resultpath)
    except OSError as e:
        print("Error: %s - %s." % (e.filename, e.strerror))
        return str(dirname(dirname(abspath(__file__)))) + "/"





class ArgparseSimulator():
    def __init__(self,
                 resultpath = "results/GenNet_experiment_2_/",
                 type = "NID",
                 layer = None
                 ):

        self.resultpath = resultpath
        self.type = type
        self.layer = layer

class ArgparseSimulator():
    def __init__(self,
                 path='/',
                 ID=999,
                 genotype_path='undefined',
                 network_name='undefined',
                 problem_type="classification",
                 wpc=1,
                 lr=0.01,
                 bs=10,
                 epochs=10,
                 L1=0.001,
                 patience = 10,
                 epoch_size = 100,
                 mixed_precision=False,
                 outfolder="undefined",
                 suffix='',
                 num_eval = 100,
                 num_sample_pat=100,
                 start_rank = 0,
                 end_rank = 0,
                 resultpath = "results/GenNet_experiment_2_/",
                 type = "NID",
                 layer = None):
        self.path = path
        self.ID = ID
        self.genotype_path = genotype_path
        self.network_name = network_name
        self.problem_type = problem_type
        self.wpc = wpc
        self.lr = lr
        self.learning_rate = lr
        self.bs = bs
        self.batch_size = bs
        self.epochs = epochs
        self.L1 = L1
        self.mixed_precision = mixed_precision
        self.out = outfolder
        self.suffix = suffix
        self.patience = patience
        self.epoch_size = epoch_size
        self.num_eval = num_eval
        self.resultpath = resultpath
        self.type = type
        self.layer = layer
        self.num_sample_pat = num_sample_pat
        self.start_rank = start_rank
        self.end_rank = end_rank
        


class TestAtoZ():       
    def test_convert(self):
        GenNet_path = get_GenNet_path()
        test1 = os.system(
            "python {}/GenNet.py convert -g {}/examples/A_to_Z/plink/"
            " -o {}/examples/A_to_Z/processed_data/"
            "/  -study_name GenNet_simulation -step all".format(GenNet_path, GenNet_path, GenNet_path) )
        assert test1 == 0
 
    def test_topology_annovar_input(self):
        GenNet_path = get_GenNet_path()
        test1 = os.system(
            "python {}/GenNet.py topology -type create_annovar_input -path {}//examples/A_to_Z/processed_data/ -study_name GenNet_simulation -out {}/examples/A_to_Z/processed_data/".format(GenNet_path, GenNet_path, GenNet_path) )
        assert test1 == 0

    def test_topology_create_gene_network(self):
        GenNet_path = get_GenNet_path()
        test1 = os.system(
            "python {}/GenNet.py topology -type create_gene_network -path {}/examples/A_to_Z/processed_data/ -out {}/examples/A_to_Z/processed_data/ -study_name GenNet_simulation".format(GenNet_path, GenNet_path, GenNet_path) )
        assert test1 == 0

    def test_train_run(self): 
        GenNet_path = get_GenNet_path()
        os.system('mkdir {}/examples/A_to_Z/new_run_folder/'.format(GenNet_path))
        os.system('cp {}/examples/A_to_Z/processed_data/topology.csv  {}/examples/A_to_Z/new_run_folder/'.format(GenNet_path, GenNet_path))
        os.system('cp {}/examples/A_to_Z/processed_data/genotype.h5  {}/examples/A_to_Z/new_run_folder/'.format(GenNet_path, GenNet_path))
        os.system('cp {}/examples/A_to_Z/run_folder/subjects.csv  {}/examples/A_to_Z/new_run_folder/'.format(GenNet_path, GenNet_path))
        test1 = os.system('python {}/GenNet.py train -path {}/examples/A_to_Z/new_run_folder/ -ID 999999997 -epochs 5'.format(GenNet_path, GenNet_path))
        assert test1 == 0
        
    def test_manhattan_plot(self):
        GenNet_path = get_GenNet_path()
        test1 = os.system(
            "python {}/GenNet.py plot -ID 999999997 -type manhattan_relative_importance".format(GenNet_path) )
        assert test1 == 0        

        
class TestTrain():
    def test_train_classification(self): 
        remove_old(999999999)
        GenNet_path = get_GenNet_path()
        value = os.system('python {}/GenNet.py train -path {}/examples/example_classification/ -ID 999999999 -epochs 2'.format(GenNet_path, GenNet_path) )
        assert value == 0
    def test_train_regression(self):
        remove_old(999999998)
        GenNet_path = get_GenNet_path()
        value = os.system('python {}/GenNet.py train -path {}/examples/example_regression/ -ID 999999998 -problem_type regression -epochs 2'.format(GenNet_path, GenNet_path) )
        assert value == 0        
    def test_train_regression_cov(self):
        remove_old(999999997)
        GenNet_path = get_GenNet_path()
        value = os.system('python {}/GenNet.py train -path {}/examples/example_regression_cov/ -ID 999999997 -problem_type regression -epochs 5'.format(GenNet_path, GenNet_path) )
        assert value == 0  
#     def test_train_regression_cov_mf(self):
#         remove_old(999999996)
#         GenNet_path = get_GenNet_path()
#         value = os.system('python {}/GenNet.py train -path {}/examples/example_regression_cov/ -ID 999999996 -problem_type regression -network_name gene_network_multiple_filters -filters 2 -epochs 5'.format(GenNet_path, GenNet_path) )
#         assert value == 0  
#     def test_train_regression_cov_mf2(self):
#         remove_old(999999995)
#         GenNet_path = get_GenNet_path()
#         value = os.system('python {}/GenNet.py train -path {}/examples/example_regression_cov/ -ID 999999995 -problem_type regression -network_name gene_network_snp_gene_filters -filters 2 -epochs 5'.format(GenNet_path, GenNet_path) )
#         assert value == 0  

        

@pytest.mark.parametrize("ID", [999999999, 999999998]) # test both regression and classification
class TestPlot():       
    def test_sunburst(self, ID):
        GenNet_path  =  get_GenNet_path()
        resultpath = GenNet_path +  "results/GenNet_experiment_" + str(ID)  + '_/'
        importance_csv = pd.read_csv(resultpath + "/connection_weights.csv", index_col=0)
        sunburst_plot(resultpath, importance_csv)
        
    def test_manhattan(self, ID):
        GenNet_path  =  get_GenNet_path()
        resultpath = GenNet_path +  "results/GenNet_experiment_" + str(ID)  + '_/'
        importance_csv = pd.read_csv(resultpath + "/connection_weights.csv", index_col=0)
        manhattan_relative_importance(resultpath, importance_csv)
        
    def test_plot_layer_weight_1(self, ID):
        GenNet_path  =  get_GenNet_path()
        resultpath = GenNet_path +  "results/GenNet_experiment_" + str(ID)  + '_/'
        importance_csv = pd.read_csv(resultpath + "/connection_weights.csv", index_col=0)
        plot_layer_weight(resultpath, importance_csv, layer=0)
    def test_plot_layer_weight_2(self, ID):
        GenNet_path  =  get_GenNet_path()
        resultpath = GenNet_path +  "results/GenNet_experiment_" + str(ID)  + '_/'
        importance_csv = pd.read_csv(resultpath + "/connection_weights.csv", index_col=0)
        plot_layer_weight(resultpath, importance_csv, layer=1)
    

class TestInterpret():
    def test_interpret_NID(self):
        GenNet_path = get_GenNet_path()
        args = ArgparseSimulator(resultpath=GenNet_path + "results/GenNet_experiment_999999999_/", type="NID")
        interpret(args)
    def test_interpret_DFIM(self):
        GenNet_path = get_GenNet_path()
        args = ArgparseSimulator(resultpath=GenNet_path + "results/GenNet_experiment_999999999_/", genotype_path = "examples/example_classification/", type="DFIM", num_eval = 100, num_sample_pat=100)
        interpret(args)

        



import argparse
import os

import h5py
import numpy as np
import pandas as pd
import tables

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Script to remove subjects from HASE hdf5 files')
    parser.add_argument("-g", required=True, type=str, help="path/paths to genotype data folder")
    parser.add_argument('-study_name', type=str, required=True, default=None, help=' Study names')
    parser.add_argument('-exclude_ids', type=str, default=None,
                        help='Table with IDs to exclude from data. Should have ID column')
    args = parser.parse_args()
    print(args)

    if args.exclude_ids is not None:
        df = pd.DataFrame.from_csv(args.snp_id_inc, index_col=None)
        print(df.head())
        if 'ID' not in df.columns:
            raise ValueError('{} table does not have ID or columns'.format(args.exclude_ids))

        df['ID'] = df.ID.astype(str)

        df_ids = pd.read_hdf(os.path.join(args.g, 'individuals', args.study_name + '.h5'), 'individuals')
        df_ids['individual'] = df_ids.individual.astype(str)

        info_index = df_ids.individual.isin(df.ID)
        remove_index = np.where(info_index == True)[0]
        keep_index = np.where(info_index == False)[0]

        if len(remove_index) == 0:
            print('There is no ids to remove!')
            exit(0)
        if len(keep_index) == len(df_ids.individual):
            print("Need to remove everybody!!! ")
            exit(0)

        individuals = df_ids.individual[~remove_index]
        chunk = pd.DataFrame.from_dict({"individual": individuals})
        chunk.to_hdf(os.path.join(args.g, 'individuals', args.study_name + '.h5'), key='individuals', format='table',
                     min_itemsize=25, complib='zlib', complevel=9)

        for g_file in os.listdir(os.path.join(args.g, 'genotype')):
            print(g_file)

            data = h5py.File(os.path.join(args.g, 'genotype', g_file), 'r')['genotype'][...]
            data = data[:, keep_index]

            h5_gen_file = tables.open_file(
                os.path.join(args.g, 'genotype', g_file), 'w', title=args.study_name)

            atom = tables.Float16Atom()
            genotype = h5_gen_file.create_carray(h5_gen_file.root, 'genotype', atom,
                                                 (data.shape),
                                                 title='Genotype',
                                                 filters=tables.Filters(complevel=9, complib='zlib'))
            genotype[:] = data
            h5_gen_file.close()



import gc
import glob
import os
import shutil
from collections import OrderedDict

import h5py
import numpy as np
import pandas as pd
from numpy import genfromtxt

from gennet.utils.hase.hdgwas.tools import Mapper


try:
    import bitarray as ba
except:
    print('bitarray failed to import this might give some trouble converting binary files')

class MINIMACPool(object):

    def __init__(self):
        self.paths = {}
        self.id = None
        self.max_index = None
        self.chunk_size = None
        self.finish = 0
        self.split_size = None

    def get_chunk(self, indices=None):

        if self.max_index is None and len(self.path) == 0:
            raise ValueError('There is no data in pool.')
        else:
            sample = h5py.File(self.paths[self.id[0]], 'r')[self.id[0]][...]
            self.max_index = np.max(sample.shape)

        print(self.finish)
        result = []
        if indices is None:
            if self.finish != self.max_index:
                n = list(range(self.finish, self.finish + self.chunk_size)) if (
                                                                                       self.finish + self.chunk_size) < self.max_index else list(
                    range(self.finish, self.max_index))
                n = np.array(n)
                n_s = np.argsort(n)
                n_s_s = np.argsort(n_s)
                self.finish = np.max(n) + 1

            else:
                return None
        else:
            n_s = np.argsort(indices)
            n_s_s = np.argsort(n_s)

        for k in self.id:
            d = h5py.File(self.paths[k], 'r')[k][list(n[n_s])]
            result.append(d[n_s_s])
        result = np.array(result)
        result = result.T
        return result

    def remove(self, type=None):
        if type is None:
            raise ValueError('Need to set type to pool remove function')
        if self.finish != self.max_index:
            print('WARNING! Removing files before fully read them!')
        print('Start to remove MINIMAChdf5 id files...')
        for i in self.id:
            os.remove(self.paths[i])
        print('Finished')

    def move(self, dst, type=None):
        if type is None:
            raise ValueError('Need to set type to pool move type')
        if self.finish != self.max_index:
            print('WARNING! Moving files before fully read them!')
        print('Start to move MINIMAChdf5 id files to id_genotype folder...')
        for i in self.id:
            shutil.move(self.paths[i], os.path.join(dst, i + '.h5'))
        print('Finished')


class PhenPool(object):

    def __init__(self, folder):
        self.paths = {}
        self.loaded = {}
        self.readed = {}
        self.inmem = 0
        self.limit = 2
        self.split_size = None
        self.folder = folder
        self.keys = self.folder.files
        self.len_dict = {k: len(self.folder.data_info[k]) for k in self.keys}

    # @timing
    def link(self, n):
        r = np.zeros((len(n), 2), dtype=np.int64)
        n = np.array(n)
        ind = 0
        for j, k in enumerate(self.keys):
            n = n - self.len_dict[k]
            index = np.where(n < 0)[0]
            if (len(index)) != 0:
                r[index, 0] = j
                r[index, 1] = n[index] + self.len_dict[k]
                n[index] = 10 ** 9  # replacing by huge number, should be bigger than actual number of phenotypes
                ind += len(index)
                if ind == len(n):
                    break
        return r

    # @timing
    def get_chunk(self, indices):
        indices = self.link(indices)
        indices = np.array(indices)
        keys, ind = np.unique(indices[:, 0], return_inverse=True)
        result = None
        # gc.collect()
        for i, k in enumerate(keys):
            r = None
            if k in self.loaded:
                n = indices[(ind == i), 1].flatten()
                r = self.loaded[self.keys[k]][n, :]
            else:
                self.folder.read(self.keys[k])
                r = self.folder._data._data
                n = indices[(ind == i), 1].flatten()
                r = r[:, n]
            l = r.shape[0]
            if result is None:
                result = np.empty((l, indices.shape[0]))
            result[:, (ind == i)] = r
        return result


class Pool(object):

    def __init__(self):
        self.paths = {}
        self.loaded = {}
        self.keys = []
        self.readed = {}
        self.inmem = 0
        self.limit = 2
        self.split_size = None

    # @timing
    def link(self, n):
        chunks = n / self.split_size
        ind = n - self.split_size * chunks
        r = np.zeros((len(n), 2), dtype=int)
        r[:, 0] = chunks
        r[:, 1] = ind
        return r

    # @timing
    def get_data(self, key, index):
        if key in self.loaded:
            r = self.loaded[key][index, :]
            self.readed[key] -= 1
            if self.readed[key] == 0:
                self.inmem -= 1
                self.loaded[key] = None
                del self.loaded[key]
                gc.collect()

        else:
            if self.inmem < self.limit:
                self.loaded[key] = h5py.File(self.paths[key], 'r')['genotype'][...]
                self.inmem += 1
                self.readed[key] = self.loaded[key].shape[0]
                r = self.loaded[key][index, :]
                self.readed[key] -= 1

            else:
                f = h5py.File(self.paths[key], 'r')
                r = f['genotype'][index, :]
                f.close()
        return r

    # @timing
    def get_chunk(self, indices, impute):
        indices = self.link(indices)
        indices = np.array(indices)
        keys, ind = np.unique(indices[:, 0], return_inverse=True)
        result = None
        for i, k in enumerate(keys):
            r = None
            if k in self.loaded:
                n = indices[(ind == i), 1].flatten()
                r = self.loaded[k][n, :]
            else:
                f = h5py.File(self.paths[k], 'r')
                n = indices[(ind == i), 1].flatten()
                r = f['genotype'][...]
                r = r[n, :]
                f.close()

            l = r.shape[1]
            if result is None:
                print(indices.shape[0], l)
                result = np.empty((indices.shape[0], l))

            result[(ind == i), :] = r
        if impute:  # default True
            if np.sum(result == 9) != 0:
                print('Imputing missing genotype to mean...')
                m = np.where(result == 9)
                result[m] = np.nan
                result[m] = np.take(np.nanmean(result, axis=1), m[0])
                # result[m]=np.nanmean(result[m[0], :], axis=1)
                print(result)

            if np.sum(np.isnan(result)) != 0:
                print('Imputing missing genotype to mean...')
                m = np.where(np.isnan(result))
                result[m] = np.take(np.nanmean(result, axis=1), m[0])
        # print result
        # result[np.where(np.isnan(result))]=np.nanmean(result[np.where(np.isnan(result))[0], :], axis=1)

        return result


class Data(object):

    def __init__(self):
        self.id = None
        self.shape = None
        self.names = None
        self.type = None
        self.format = None
        self._data = None
        self.chunk_size = None
        self.processed = 0
        self.start = None
        self.finish = None
        self.filename = None

    def test(self):
        if isinstance(self.id, type(None)) or isinstance(self.names, type(None)) or isinstance(self._data, type(None)):
            raise ValueError('Read data first!')
        else:
            N, M = self._data.shape
            uniq_id = np.unique(self.id)
            if len(uniq_id) != N:
                print(uniq_id)
                raise ValueError('Should be uniq ID, but there are {} only uniq from {}'.format(len(uniq_id), N))
            uniq_names = np.unique(self.names)
            if len(uniq_names) != M:
                print(uniq_names)
                raise ValueError('Should be uniq NAMES, but there are {} only uniq from {}'.format(len(uniq_id), M))

    def get_id(self):
        if not isinstance(self.id, type(None)):
            return np.copy(np.array(self.id))
        else:
            raise ValueError('No id defined!')

    def get_names(self):
        if not isinstance(self.names, type(None)):
            return (np.array(self.names))
        else:
            raise ValueError('No names defined!')

    def get_next(self, index=None, select_columns=None):

        if isinstance(self._data, type(None)):
            return None

        if self.processed == self.shape[1]:
            return None
        else:
            if select_columns is not None:
                if not isinstance(index, type(None)):
                    return self._data[index, select_columns]
                else:
                    return self._data[:, select_columns]
            else:
                start = self.processed
                finish = self.processed + self.chunk_size if (self.processed + self.chunk_size) <= self.shape[1] else \
                    self.shape[1]
                self.processed = finish
                if not isinstance(index, type(None)):
                    self.start = start
                    self.finish = finish
                    return self._data[index, start:finish]
                else:
                    self.start = start
                    self.finish = finish
                    return self._data[:, start:finish]


class Hdf5Data(Data):
    def __init__(self, data_path, name, type='PLINK'):
        super(Hdf5Data, self).__init__()
        self.name = name
        self.id = np.array(
            pd.read_hdf(os.path.join(data_path, 'individuals', self.name + '.h5'), 'individuals').individual.tolist())
        if "MAPPER" in os.environ:
            print('Reading ID from probes....')
            self.names = pd.read_hdf(os.path.join(data_path, 'probes', self.name + '.h5'), 'probes',
                                     where='columns=[ID]').ID
            self.shape = (len(self.id), len(self.names))
        else:
            print('Use ID from mapper')
            self.names = pd.HDFStore(os.path.join(data_path, 'probes', self.name + '.h5'), 'r')
            self.shape = (len(self.id), self.names.get_storer('probes').nrows)
        if type == 'PLINK':
            l = os.listdir(os.path.join(data_path, 'genotype'))
            a, b = h5py.File(os.path.join(data_path, 'genotype', l[0]), 'r')['genotype'][...].shape
            try:
                c, d = h5py.File(os.path.join(data_path, 'genotype', l[1]), 'r')['genotype'][...].shape
                self.gen_split_size = np.max((a, c))  # TODO (middle) not efficient
            except:
                self.gen_split_size = a

        print(('There are %d ids' % (self.shape[0])))


class ParData(Data):
    def __init__(self):
        super(ParData, self).__init__()
        self.a_inv = None
        self.a_cov = None
        self.b_cov = None
        self.C = None
        self.a_test = None
        self.metadata = None
        self.b4 = None

    def check(self):  # TODO (middle) place to implement the check of analysis protocol and log summary from PD
        print(('Number of subjects {}'.format(int(self.a_cov[0, 0]))))

    def get(self, gen_order=None, phen_order=None, cov_order=None):

        if gen_order is None or phen_order is None or cov_order is None:
            raise ValueError('PD order is not define!')
        if isinstance(self.a_inv, type(None)):
            return self.a_test[np.ix_(gen_order, np.append(cov_order, self.a_test.shape[1] - 1))], self.b_cov[
                np.ix_(cov_order, phen_order)], \
                   self.C[phen_order], self.a_cov[np.ix_(cov_order, cov_order)]
        else:
            return self.a_inv[np.ix_(gen_order, np.append(cov_order, self.a_test.shape[1] - 1))], self.b_cov[
                np.ix_(cov_order, phen_order)], \
                   self.C[phen_order], self.a_cov[np.ix_(cov_order, cov_order)]


class MetaPhenotype(object):

    def __init__(self, phen, protocol=None, include=None, exclude=None):
        # @timing
        def _check(map, keys, index=None):
            values = np.array(list(map.dic.values()))
            result = {}
            r = (values == -1).any(axis=1)
            if index is not None:
                r[np.where(index == False)[0]] = True
            if np.sum(r) == values.shape[0]:
                raise ValueError('There is no common names between studies')
            for i, k in enumerate(keys):
                result[k] = values[~r, i]
            return result, np.array(list(map.dic.keys()))[~r]

        self.chunk_size = 10000
        self.exclude = None
        self.include = None
        self.name = None
        self.mapper = Mapper()
        self.keys = []
        self.keep_index = None
        if include is not None:
            self.include = pd.DataFrame.from_csv(include, index_col=None)
            print('Include:')
            print(self.include.head())
            if 'ID' not in self.include.columns:
                raise ValueError('{} table does not have ID column for phenotypes'.format(include))
        if exclude is not None:
            self.exclude = pd.DataFrame.from_csv(exclude, index_col=None)
            print('Exclude:')
            print(self.exclude.head())
            if 'ID' not in self.exclude.columns:
                raise ValueError('{} table does not have ID column for phenotypes'.format(exclude))
        if protocol is None:
            for i, k in enumerate(phen):
                phen_names = []
                if i == 0:
                    for j in k.folder.files:
                        if j != 'info_dic.npy':
                            phen_names = phen_names + list(k.folder.data_info[j])
                    self.mapper.fill(phen_names, i, reference=False)
                else:
                    for j in k.folder.files:
                        if j != 'info_dic.npy':
                            phen_names = phen_names + list(k.folder.data_info[j])
                    self.mapper.push(phen_names, name=i, new_id=False)

                self.keys.append(i)
            if self.exclude is not None or self.include is not None:
                phen_names = pd.Series(list(self.mapper.dic.keys()))
                phen_names = phen_names[phen_names.isin(self.include.ID)] if self.include is not None else phen_names
                phen_names = phen_names[~phen_names.isin(self.exclude.ID)] if self.exclude is not None else phen_names
                self.keep_index = pd.Series(list(self.mapper.dic.keys())).isin(phen_names)
            print(np.sum(self.keep_index))
            self.order, self.phen_names = _check(self.mapper, self.keys, index=self.keep_index)
            self.n_phenotypes = len(self.order[self.keys[0]])
            print(('Loaded {} common phenotypes for meta-analysis'.format(self.n_phenotypes)))
            self.processed = 0
        else:
            if not protocol.enable:
                protocol.parse()

        self.pool = {i: PhenPool(j.folder) for i, j in enumerate(phen)}

    def get(self):
        if self.processed == self.n_phenotypes:
            return None, None
        else:
            start = self.processed
            finish = self.processed + self.chunk_size if (
                                                                 self.processed + self.chunk_size) <= self.n_phenotypes else self.n_phenotypes
            self.processed = finish
            phenotype = np.zeros(
                (np.sum([len(self.pool[i].folder._data.id) for i in self.pool]), len(list(range(start, finish)))))
        for i, j in enumerate(self.keys):

            if i == 0:
                a, b = 0, self.pool[j].folder._data.id.shape[0]
                phenotype[a:b, :] = self.pool[j].get_chunk(self.order[j][start:finish])
                a = b
                N = np.random.randint(0, phenotype.shape[1], 10)
                print(phenotype.mean(axis=0)[N])
            else:
                ph_tmp = self.pool[j].get_chunk(self.order[j][start:finish])
                print(self.pool[j].folder.path)
                print(ph_tmp.mean(axis=0)[N])
                b += ph_tmp.shape[0]
                phenotype[a:b, :] = ph_tmp
                a = b

        return phenotype, self.phen_names[start:finish]


class MetaParData(object):

    def __init__(self, pd, study_names, protocol=None):
        # @timing
        def _check(map, keys):
            values = np.array(list(map.dic.values()))
            result = {}
            r = (values == -1).any(axis=1)
            if np.sum(r) == values.shape[0]:
                raise ValueError('There is no common names between studies')
            for i, k in enumerate(keys):
                result[k] = values[~r, i]
            return result

        self.name = None
        self.study_names = study_names
        self.phen_mapper = Mapper()
        self.cov_mapper = Mapper()
        self.covariates = OrderedDict()
        self.pd = OrderedDict()
        for i in pd:
            self.pd[i.folder.name] = i
        keys = []
        if protocol is None:
            for i, k in enumerate(pd):
                self.covariates[k] = [n.split(self.study_names[i] + '_')[1] for n in k.folder._data.metadata['names']]
                if i == 0:
                    self.phen_mapper.fill(k.folder._data.metadata['phenotype'], k.folder.name, reference=False)
                    self.cov_mapper.fill(self.covariates[k], k.folder.name, reference=False)
                else:
                    self.phen_mapper.push(k.folder._data.metadata['phenotype'], name=k.folder.name, new_id=False)
                    self.cov_mapper.push(self.covariates[k], name=k.folder.name, new_id=False)
                keys.append(k.folder.name)

            self.phen_order = _check(self.phen_mapper, keys)
            self.cov_order = _check(self.cov_mapper, keys)

        if protocol is not None:
            if not protocol.enable:
                protocol.parse()

    def check_pd(self, old, new):

        np.set_printoptions(precision=3, suppress=True)
        print("*******PD CHECK*********")
        print('A covariates...')
        print(old[3] / old[3][0, 0])
        print(new[3] / new[3][0, 0])
        print('FREQ TESTs')
        N = np.random.randint(0, np.min((old[0].shape[0], new[0].shape[0])), 10)
        print(np.array(old[0][:, 0] / old[3][0, 0] / 2)[N])
        print(np.array(new[0][:, 0] / new[3][0, 0] / 2)[N])
        print('FREQ PHENO')
        M = np.random.randint(0, np.min((old[1].shape[1], new[1].shape[1])), 10)
        print(np.array(old[1][0, :] / old[3][0, 0])[M])
        print(np.array(new[1][0, :] / new[3][0, 0])[M])
        print("****************")

    # @timing
    def check_maf(self, SNPs_index):
        maf = np.zeros((SNPs_index[0].shape[0], len(self.pd)))

        for i, j in enumerate(self.pd):
            maf[:, i] = np.array(self.pd[j].folder._data.metadata['MAF'])[SNPs_index[i].astype(np.int64)]

        if (np.std(maf, axis=1) > 0.1).any():
            raise ValueError('MAF is not consistent between PD data!')

    def get(self, SNPs_index=None, B4=False, regression_model=None, random_effect_intercept=False):

        if self.pd is None:
            raise ValueError('Data not defined!')
        k = list(self.pd.keys())
        if SNPs_index is not None:
            if len(self.pd) != len(SNPs_index):
                raise ValueError(
                    'There are not equal number od PD and SNPs indexes {}!={}'.format(len(self.pd), len(SNPs_index)))
            a_test, b_cov, C, a_cov = self.pd[k[0]].get(gen_order=SNPs_index[0], phen_order=self.phen_order[k[0]],
                                                        cov_order=self.cov_order[k[0]])
            if random_effect_intercept and len(self.pd) > 1:  # TODO (high)
                a_test_effect = a_test[:, 0:1]
                b_cov_effect = b_cov[0:1, :]
                a_cov_effect = a_cov[0:1, :]
        else:
            raise ValueError('Indexes are not defined!')

        if B4:
            b4 = self.pd[k[0]].folder._data.b4[SNPs_index[0], :]
            b4 = b4[:, self.phen_order[k[0]]]

        # self.check_maf(SNPs_index)

        for i in range(1, len(self.pd)):
            a, b, c, a_c = self.pd[k[i]].get(gen_order=SNPs_index[i], phen_order=self.phen_order[k[i]],
                                             cov_order=self.cov_order[k[i]])
            self.check_pd([a_test, b_cov, C, a_cov], [a, b, c, a_c])
            a_test = a_test + a
            b_cov = b_cov + b
            C = C + c
            a_cov = a_cov + a_c
            if random_effect_intercept and i < (len(self.pd) - 1):
                a_test_effect = np.hstack((a_test_effect, a[:, 0:1]))
                b_cov_effect = np.vstack((b_cov_effect, b[0:1, :]))
                a_cov_effect = np.vstack((a_cov_effect, a_c[0:1, :]))
            if B4:
                b4_tmp = self.pd[k[i]].folder._data.b4[SNPs_index[i], :]
                b4 = b4 + b4_tmp[:, self.phen_order[k[i]]]

        if random_effect_intercept and len(self.pd) > 1:
            a_cov_I = np.zeros((len(self.pd) - 1, len(self.pd) - 1))
            np.fill_diagonal(a_cov_I, a_cov_effect[:, 0])
            a_effect = np.hstack((a_cov_I, a_cov_effect))
            a_test = np.hstack((a_test_effect, a_test))
            b_cov = np.vstack((b_cov_effect, b_cov))
            a_cov = np.vstack((a_cov_effect, a_cov))
            a_cov = np.hstack((a_effect.T, a_cov))

        if B4:
            return a_test, b_cov, C, a_cov, b4
        else:
            return a_test, b_cov, C, a_cov

    def maf_pard(self, SNPs_index=None):

        samples = 0
        maf = np.zeros(len(SNPs_index[0]))

        for j, i in enumerate(self.pd):
            n = len(self.pd[i].folder._data.metadata['id'])
            samples += n
            maf = maf + n * np.array(self.pd[i].folder._data.metadata['MAF'])[SNPs_index[j]]
        maf = maf / np.float(samples)
        return maf

    def get_n_id(self):
        return np.sum([len(self.pd[i].folder._data.metadata['id']) for i in self.pd])

    def get_phenotype_names(self):
        n = {}
        for i in self.pd:
            n[i] = self.pd[i].folder._data.metadata['phenotype']


class Folder(object):

    def __init__(self, path):
        self.path = path
        if not os.path.isdir(self.path):
            raise ValueError('{} is not a folder'.format(self.path))
        self.format = None
        self.protected = []
        self.n_files = None
        self.read_files = {}
        self.files = None
        self.processed = 0
        self._data = Data()
        self._scan()
        self.folder_cache = {}
        self.folder_cache_flag = True
        self.cache_buffer_size = 10
        self._id = None

    def _scan(self):
        l = os.listdir(self.path)
        if len(l) == 0:
            raise ValueError('There is no files in {}'.format(self.path))
        exts = [os.path.splitext(i)[1] for i in l]
        s = set(exts)
        if not len(s) <= 1:
            if s == {'.fam', '.bed', '.bim'}:
                self.format = 'PLINK'
            else:
                print("path", self.path)
                print(l)
                raise ValueError('There are different data format {} in {} folder'.format(set(exts), self.path))
        else:
            N = len(l)
            self.n_files = N
            self.format = exts[0]
            self.files = l

    def get_next(self, **kwargs):
        d = self._data.get_next(**kwargs)
        if isinstance(d, type(None)):
            file = next(self)
            if isinstance(file, type(None)):
                return None
            self.read(file)
            return self._data.get_next(**kwargs)
        else:
            return d

    def read(self, *args, **kwargs):
        pass

    def __next__(self):
        while True:
            try:
                if self.processed == self.n_files:
                    print('read all', self.path)
                    return None
                f = self.files[self.processed]
                self.processed += 1
                if f not in self.protected:
                    self.read_files[f] = 0
                    break
                else:
                    self.read_files[f] = 1
            except:
                return None
        return f

    def get(self, **kwargs):
        return self._data.get(**kwargs)


class HDF5Folder(Folder):

    def __init__(self, path, name):
        super(HDF5Folder, self).__init__(path)

    def _scan(self):
        pass

    def get_next(self, **kwargs):
        pass

    def read(self, file, index=None, MAF=None):
        pass

    def get(self, index):
        pass


class PLINKHDF5Folder(HDF5Folder):

    def __init__(self, path, name):
        super(HDF5Folder, self).__init__(path)
        self.name = name
        self.path = path
        self.format = '.h5'
        self.pool = Pool()
        self.iter = -1
        if not os.path.isdir(os.path.join(self.path, 'genotype')) or not os.path.isdir(
                os.path.join(self.path, 'probes')) or not os.path.isdir(os.path.join(self.path, 'individuals')):
            raise ValueError('in genotype folder should be /genotype; /probes; /individuals folders')

        self.f_names = os.listdir(os.path.join(self.path, 'genotype'))  # TODO (high) remove this extra attribute
        self.files = self.f_names
        self.n_files = len(self.f_names)
        self.processed = -1
        self.pool.paths = {i: os.path.join(self.path, 'genotype', str(i) + '_' + self.name + '.h5') for i in
                           range(self.n_files)}

        self._data = Hdf5Data(self.path, self.name)
        self._id = self._data.id
        self.pool.split_size = self._data.gen_split_size

    def _scan(self):
        pass

    def get_next(self, **kwargs):

        self.iter += 1
        if self.iter == self.n_files:
            return None
        file = os.path.join(self.path, 'genotype', str(self.iter) + '_' + self.name + '.h5')
        return self.read(file, **kwargs)

    def read(self, file, index=None, MAF=None):
        print('reading file {}'.format(file))

        f = h5py.File(file, 'r')

        d = f['genotype'][...]

        if isinstance(index, type(None)):  # TODO (mid) check
            return d
        else:
            return d[:, index]

    def get(self, index, impute=True):
        self.processed += len(index)
        result = []
        if True:
            # if self.pool.inmem==self.pool.limit: #TODO (high) check the bellow methods, remove get_data from pool?
            result = self.pool.get_chunk(index, impute)
        else:
            result = [self.pool.get_data(i, j) for i, j in index]
        return np.array(result)

    def get_info(self, file):
        info = {}
        f = h5py.File(file, 'r')
        info['shape'] = f['genotype'].shape
        f.close()
        return info


class MINIMACHDF5Folder(HDF5Folder):

    def __init__(self, path, name):
        super(HDF5Folder, self).__init__(path)
        self.name = name
        self.path = path
        self.format = '.h5'
        self.pool = MINIMACPool()
        self.pool.path = self.path
        self.f_names = os.listdir(os.path.join(self.path, 'genotype'))
        self.n_files = len(self.f_names)
        self._data = Hdf5Data(self.path, self.name, type='MINIMAC')
        self._id = self._data.id
        self.pool.id = self._id
        self.pool.paths = {i: os.path.join(self.path, 'genotype', i + '.h5') for i in self._id}

    def _scan(self):
        pass

    def get_next(self, **kwargs):
        return self.pool.get_chunk()

    def get(self, index):
        return self.pool.get_chunk(indices=index)

    def read(self, file, index=None, MAF=None):
        pass

    def get(self, index):
        pass


class CSVFolder(Folder):

    def __init__(self, path):
        super(CSVFolder, self).__init__(path)
        try:
            self.read(next(self))
        except Exception as e:
            raise (e)
        self.data_info = {}
        for i in self.files:
            for j in ['\t', ' ']:
                df = pd.read_csv(os.path.join(self.path, i), sep=j, index_col=None)
                if df.shape[1] > 1:
                    break
            self.data_info[i] = np.array(df.columns[1:])

    def read(self, file):

        if self.folder_cache_flag and self.folder_cache.get(file) is not None and len(
                self.folder_cache) < self.cache_buffer_size:
            self._data = self.folder_cache[file]
            self._data.processed = 0
            print(('There are samples and columns {} in cache {}'.format(self._data.shape, file)))

        else:
            print('reading file {}'.format(file))
            for i in ['\t', ' ']:
                df = pd.read_csv(os.path.join(self.path, file), sep=i, index_col=None)
                if df.shape[1] > 1:
                    break
            else:
                raise ValueError(
                    'Cant read {} file; default settings: index=None, header=True; sep=tab or space '.format(file))

            self._data = Data()
            self._data.chunk_size = 1000
            self._data._data = df[df.columns[1:]].as_matrix()
            self._data.id = np.array(df[df.columns[0]])
            if self._id is None:
                self._id = self._data.id
            else:
                if (
                        self._id != self._data.id).any():  # TODO (middle) check for not overlapped id (AttributeError: 'bool' object has no attribute 'any')
                    raise ValueError('id in {} different from previous!'.format(file))
            self._data.type = np.ndarray
            self._data.names = np.array(df.columns[1:])
            self._data.shape = self._data._data.shape
            self._data.format = 'csv'
            self._data.filename = file

            self._data.test()  # check the dim of data

            if self.folder_cache_flag and len(self.folder_cache) < self.cache_buffer_size:
                self.folder_cache[file] = self._data

            print(('There are %d ids and %d columns ' % (self._data.shape)))


class NPFolder(Folder):

    def __init__(self, path, protected=[]):
        super(NPFolder, self).__init__(path)
        self._data = Data()
        self._data.chunk_size = 1000
        self._data.type = np.ndarray
        self._data.format = 'npy'
        self.protected = protected

        try:
            file_path = os.path.dirname(path)
            self.data_info = np.load(os.path.join(file_path, 'info_dic.npy')).item()
            self._data.id = np.array(self.data_info['id'])
            self.files = [k for k in list(self.data_info.keys()) if k != 'id']

        except Exception as e:
            raise ValueError('in directory {} should be data info file info_dic.npy!'.format(self.path) + str(e))

        try:
            self.read(next(self))
        except Exception as e:
            raise ValueError("Failed to init NPFolder;" + str(e))

    # @timing
    def read(self, file):

        if self.folder_cache_flag and self.folder_cache.get(file) is not None and len(
                self.folder_cache) < self.cache_buffer_size:
            print('start cache {}'.format(file))
            self._data._data = self.folder_cache[file]
            self._data.processed = 0
            self._data.type = np.ndarray
            self._data.shape = self._data._data.shape
            self._data.format = 'npy'
            self._data.filename = file
            self._data.names = np.array(self.data_info[file])
        # print 'end cache'
        else:
            # print 'reading file {}'.format(file)
            d = np.load(os.path.join(self.path, file))
            if not isinstance(d, np.ndarray):
                raise ValueError('File {} saved not as numpy array'.format(file))
            self._data.processed = 0
            self._data._data = d
            self._data.type = np.ndarray
            self._data.shape = d.shape
            self._data.format = 'npy'
            self._data.filename = file
            self._data.names = np.array(self.data_info[file])

            self._data.test()  # check the dim of data
            if self.folder_cache_flag and len(self.folder_cache) < self.cache_buffer_size:
                self.folder_cache[file] = self._data._data


class PDFolder(Folder):

    def __init__(self, path, name):
        super(PDFolder, self).__init__(path)
        self.name = name
        files = [self.name + '_a_cov.npy', self.name + '_b_cov.npy', self.name + '_C.npy', self.name + '_metadata.npy']
        for i in files:
            if i not in self.files:
                raise ValueError('There is not {} file in directory {}'.format(i, self.path))
        self._data = ParData()
        self.loaded = False

    def scan(self):
        pass

    def get_next(self, **kwargs):
        pass

    def get(self, **kwargs):
        return self._data.get(**kwargs)

    def read(self, file):
        pass

    def load(self):

        if self.name + '_b4.npy' in self.files:
            self._data.b4 = np.load(os.path.join(self.path, self.name + '_b4.npy'))

        self._data.a_cov = np.load(os.path.join(self.path, self.name + '_a_cov.npy'))
        self._data.b_cov = np.load(os.path.join(self.path, self.name + '_b_cov.npy'))
        self._data.C = np.load(os.path.join(self.path, self.name + '_C.npy'))
        self._data.metadata = np.load(os.path.join(self.path, self.name + '_metadata.npy')).item()

        if self.name + '_a_inv.npy' in self.files:  # TODO (low) current version does not save inv matrix
            self._data.a_inv = np.load(os.path.join(self.path, self.name + '_a_inv.npy'))

        elif self.name + '_a_test.npy' in self.files:
            self._data.a_test = np.load(os.path.join(self.path, self.name + '_a_test.npy'))

        else:
            raise ValueError('There is not a_inv.npy or a_test.npy file in directory {}'.format(self.path))

        self.loaded = True

    def summary(self):  # TODO (middle) write function
        if not self.loaded:
            raise ValueError('call summary PD before load data!')


class PLINKFolder(Folder):

    def __init__(self, path):
        super(PLINKFolder, self).__init__(path)
        self.scan()
        self.names = None
        self.bim = None
        self.bed = None
        self.fam = None
        self.N_probes = 0
        self.n_probes = 0
        self.n_probes_dic = {}
        self.n_ind_dic = {}
        self.n_ind = 0
        self.N_ind = 0
        self._currentSNP = 0

        self._bedcode = {
            2: ba.bitarray('11'),
            9: ba.bitarray('10'),  # TODO (high) NA data handle
            1: ba.bitarray('01'),
            0: ba.bitarray('00')
        }

        self.scan()

    def scan(self):

        l = os.listdir(self.path)
        self.names = np.unique([os.path.splitext(i)[0] for i in l])

        for i in self.names:
            if i + '.fam' not in l or i + '.bed' not in l or i + '.bim' not in l:
                raise ValueError('There are not all plink files for {} in {} folder'.format(i, self.path))

        self.n_files = len(l)
        self.format = 'PLINK'
        self.files = self.names

    def get_fam(self):
        """Read the FAM file to get information about individuals
        Family ID
        Individual ID
        Paternal ID
        Maternal ID
        Sex (1=male; 2=female; other=unknown)
        Label - phenotype should be in separate file
        """

        individuals = None

        for i in self.names:
            try:
                ind = genfromtxt(open(os.path.join(self.path, i + '.fam'), 'r'), delimiter=' ',
                             dtype={'names': ['family', 'individual', 'paternal', 'maternal', 'sex', 'label'],
                                    'formats': ['S10', 'S16', int, int, int, int]})
            except:
                ind = genfromtxt(open(os.path.join(self.path, i + '.fam'), 'r'), delimiter='\t',
                             dtype={'names': ['family', 'individual', 'paternal', 'maternal', 'sex', 'label'],
                                    'formats': ['S10', 'S16', int, int, int, int]})

            if isinstance(individuals, type(None)):
                individuals = ind

            else:
                individuals = np.append(individuals, ind)
            self.n_ind_dic[i] = len(ind)
        self.N_ind = len(individuals)
        print(('Number of Individuals: %d' % self.N_ind))

        return individuals

    def get_bim(self, chunk_size):

        if isinstance(self.bim, type(None)):
            file = next(self)
            self.read_bim(file)
        try:
            d = self.bim.get_chunk(chunk_size)
            return d
        except:
            file = next(self)
            if isinstance(file, type(None)):
                return None
            self.read_bim(file)
            d = self.bim.get_chunk(chunk_size)
            return d

    def read_bim(self, file):

        # bim_lines= Popen(['wc','-l',os.path.join(self.path,file +'.bim')], stdout=PIPE)
        # bim_lines_com = bim_lines.communicate()
        # print("------------------------------------------")
        # print(bim_lines_com[0])
        # N=int(bim_lines_com[0].split(' ')[0])
        N = int(sum(1 for line in open(os.path.join(self.path, file + '.bim'))))

        print(('Number of Probes {} in {}'.format(N, file + '.bim')))
        self.N_probes += N

        self.n_probes_dic[file] = N

        self.bim = pd.read_table(os.path.join(self.path, file + '.bim'), sep='\t', header=None,
                                 names=['CHR', 'ID', 'distance', 'bp', 'allele1', 'allele2'],
                                 iterator=True)

    def get_bed(self, chunk_size):

        d = self.nextSNPs(chunk_size)

        if isinstance(d, type(None)):
            file = next(self)
            if isinstance(file, type(None)):
                return None
            else:
                self.read_bed(file)
                return self.get_bed(chunk_size)
        else:
            if d.shape[0] == chunk_size:
                return d
            else:
                d_append = self.get_bed(chunk_size - d.shape[0])
                if isinstance(d_append, type(None)):
                    return d
                else:
                    return np.vstack((d, d_append))

    def read_bed(self, file):

        self._currentSNP = 0
        self.n_ind = self.n_ind_dic[file]
        self.n_probes = self.n_probes_dic[file]
        n = self.n_ind
        self.bed = open(os.path.join(self.path, file + '.bed'), 'rb')
        magicNumber = ba.bitarray(endian="little")
        magicNumber.fromfile(self.bed, 2)
        bedMode = ba.bitarray(endian="little")
        bedMode.fromfile(self.bed, 1)
        e = (4 - n % 4) if n % 4 != 0 else 0
        nru = self.n_ind + e
        self.nru = nru
        # check magic number
        if magicNumber != ba.bitarray('0011011011011000'):
            raise IOError("Magic number from Plink .bed file not recognized")

        if bedMode != ba.bitarray('10000000'):
            raise IOError("Plink .bed file must be in default SNP-major mode")

    def nextSNPs(self, b):
        '''
        Unpacks the binary array of genotypes and returns an n x b matrix of floats of
        normalized genotypes for the next b SNPs, where n := number of samples.

        Parameters
        ----------
        b : int
            Number of SNPs to return.
        Returns
        -------
        X : np.array with dtype float64 with shape (n, b), where n := number of samples

        '''
        if self._currentSNP == self.n_probes:
            return None
        if self._currentSNP + b > self.n_probes:
            b = (self.n_probes - self._currentSNP)

        print(('next {} SNPs, from {}, need to convert {}'.format(b, self.n_probes,
                                                                  (self.n_probes - self._currentSNP - b))))

        c = self._currentSNP
        n = self.n_ind
        nru = self.nru
        slice = ba.bitarray(endian="little")

        bit_number = ((2 * (c + b) * nru) - (2 * c * nru)) // 8

        slice.fromfile(self.bed, bit_number)

        X = np.array(slice.decode(self._bedcode), dtype="float64").reshape((b, nru)).T
        X = X[0:n, :]

        self._currentSNP += b
        gc.collect()
        return X.T

    def get_next(self, **kwargs):
        pass

    def read(self):
        pass


class MINIMACFolder(Folder):

    def __init__(self, path):
        super(MINIMACFolder, self).__init__(path)
        self.scan()

    def scan(self):
        if len(glob.glob(self.path + '*dose.gz')) == 0 or len(glob.glob(self.path + '*info.gz')) == 0:
            raise ValueError(
                'There is no dose.gz or info.gz files in {}, check folder or compression flag'.format(self.path))

    def get_next(self, **kwargs):
        pass


class VCFFolder(Folder):

    def __init__(self, path):
        super(VCFFolder, self).__init__(path)
        self.scan()

    def scan(self):  # TODO (middle) add checking for correct VCF format (? checkVCF.py) before start shell scripts
        pass

    def get_next(self, **kwargs):
        pass


class Reader(object):

    def __init__(self, name):
        self.name = name
        self.path = None
        self._data = None
        self.ext = ['.npy', '.csv', '.txt', '.h5', 'PLINK', '.gz', 'VCF']
        self.pool = None
        self.folder = None
        self.processed = 0
        self.kwargs = None
        self.permutation = False

    def start(self, path, **kwargs):

        self.kwargs = kwargs
        if path is None:
            raise ValueError('Not defined path for {}'.format(self.name))
        if os.path.isdir(path):

            self.folder = Folder(path)
            self.path = path

            self.format = self.folder.format

            if kwargs.get('vcf', 0) and kwargs['vcf']:
                self.format = 'VCF'

            if self.name == 'genotype' and self.format != 'PLINK' and self.format != '.gz' and self.format != 'VCF':
                self.format = '.h5'

            if self.format not in self.ext:
                raise ValueError(' {} is not supported format (only {} supported)!'.format(self.format, self.ext))

            elif self.format == '.npy' and self.name != 'partial':
                self.folder = NPFolder(path, protected=['info_dic.npy'])
            # self.folder.protected=['info_dic.npy']

            elif self.format == '.npy' and self.name == 'partial':
                self.folder = PDFolder(path, name=kwargs['study_name'])


            elif self.format in ['.csv', '.txt']:
                self.folder = CSVFolder(path)

            elif self.format == '.h5' and self.name == 'genotype':
                self.folder = PLINKHDF5Folder(path, kwargs['study_name'])

            elif self.format == 'PLINK':
                self.folder = PLINKFolder(path)

            elif self.format == '.gz':
                self.folder = MINIMACFolder(path)
                self.format = "MINIMAC"

            elif self.format == 'VCF':
                self.folder = VCFFolder(path)

            elif self.format == '.h5' and self.name != 'genotype':
                raise ValueError('hdf5 format implemented only for converted from genotype data...Sorry')

            elif self.format != '.h5' and self.name == 'genotype':
                raise ValueError('genotype data should be in hdf5 format')

            self._get = self.folder.get_next

        else:
            raise ValueError('{} is not a directory'.format(path))

    def get(self, *args, **kwargs):
        if self.format != '.h5' and self.name != 'partial':
            raise ValueError()
        return self.folder.get(*args, **kwargs)

    def get_next(self, **kwargs):
        return self._get(**kwargs)



import os
import pickle as pickle

import numpy as np


class HashTablePOS(object):
    def __init__(self):
        self.hash_dic = {i: {} for i in range(23)}
        self.protected = False
        self.data_frame = None
        self.collisions = {i: [] for i in range(23)}
        self.type = 'POS'
        self.n_keys = 0
        self._print = False

    def fill_id(self, id):
        if self.hash_dic[id.CHR].get(id.bp) is not None:
            if isinstance(self.hash_dic[id.CHR][id.bp], tuple):
                if self._print:
                    print((
                        "Collision! You are trying to fill {} again with A1_{}/A2_{},"
                        " it is already inserted with A1_{}/A2_{}. "
                        "".format(
                            "CHR " + str(id.CHR) + " bp " + str(id.bp),
                            id.allele1, id.allele2,
                            self.hash_dic[id.CHR][id.bp][1], self.hash_dic[id.CHR][id.bp][2])
                    ))

                self.hash_dic[id.CHR][id.bp] = [self.hash_dic[id.CHR][id.bp], (self.n_keys, id.allele1, id.allele2)]
                self.n_keys += 1
            else:
                self.hash_dic[id.CHR][id.bp].append((self.n_keys, id.allele1, id.allele2))
                self.n_keys += 1
            self.collisions[id.CHR].append(id.bp)
        else:
            self.hash_dic[id.CHR][id.bp] = (self.n_keys, id.allele1, id.allele2)
            self.n_keys += 1

    def fill(self, data_frame):
        self.data_frame = data_frame
        if self.data_frame is None:
            raise ValueError('Genotype data frame is not defined!')
        for i in self.data_frame.iterrows():
            self.n_keys += 1
            if self.hash_dic[i[1].CHR].get(i[1].bp) is not None:
                if isinstance(self.hash_dic[i[1].CHR][i[1].bp], tuple):
                    if self._print:
                        print((
                            "Collision! You are trying to fill {} again with A1_{}/A2_{},"
                            " it is already inserted with A1_{}/A2_{}. "
                            "".format(
                                "CHR " + str(i[1].CHR) + " bp " + str(i[1].bp),
                                i[1].allele1, i[1].allele2,
                                self.hash_dic[i[1].CHR][i[1].bp][1], self.hash_dic[i[1].CHR][i[1].bp][2])
                        ))

                    self.hash_dic[i[1].CHR][i[1].bp] = [self.hash_dic[i[1].CHR][i[1].bp],
                                                        (i[0], i[1].allele1, i[1].allele2)]
                else:
                    self.hash_dic[i[1].CHR][i[1].bp].append((i[0], i[1].allele1, i[1].allele2))
                self.collisions[i[1].CHR].append(i[1].bp)
            else:
                self.hash_dic[i[1].CHR][i[1].bp] = (i[0], i[1].allele1, i[1].allele2)
        self.protected = True

    def get_index(self, chr, bp):

        r = self.hash_dic[chr].get(bp)
        if r is None:
            return -1, 'R', 'R'
        else:
            return r

    def get_map(self, info):

        def _map(r, a1, a2):
            ind, A1, A2 = r
            if ind == -1:
                result = [-1, 0]
            elif A1 == a1 and A2 == a2:
                result = [ind, 1]
            elif A1 == a2 and A2 == a1:
                result = [ind, -1]
            else:
                result = [-1, 0]
            return result

        r = self.get_index(info.CHR, info.bp)
        if not isinstance(r[0], tuple):
            return _map(r, info.allele1, info.allele2)
        else:
            for i in r:
                m = _map(i, info.allele1, info.allele2)
                if m[0] != -1:
                    return m
            return [-1, 0]


class HashTableRSID(object):

    def __init__(self):
        self.hash_table = np.zeros([250 * 10 ** 6])
        self.list = []
        self.hash_dic = {}
        self.protected = False
        self.type = 'RSID'
        self.n_keys = 0

    def fill_id(self, id):
        # print id
        if not self.protected:
            # self.n_keys+=1
            try:
                if id.ID[:2] == 'rs':
                    m = int(id.ID[2:])
                    if self.hash_table[m] != 0:
                        print('collision {}!'.format(id))
                        if isinstance(self.list[self.hash_table[m]], tuple):
                            self.list[self.hash_table[m]] = [self.list[self.hash_table[m]],
                                                             (self.n_keys, id['allele1'], id['allele2'])]
                            self.n_keys += 1
                        else:
                            self.list[self.hash_table[m]].append((self.n_keys, id['allele1'], id['allele2']))
                            self.n_keys += 1
                    else:
                        self.list.append((self.n_keys, id['allele1'], id['allele2']))
                        self.hash_table[m] = len(self.list)
                        self.n_keys += 1
                else:
                    if self.hash_dic.get(id.ID) is None:
                        self.hash_dic[id.ID] = (self.n_keys, id['allele1'], id['allele2'])
                        self.n_keys += 1
                    else:
                        if isinstance(self.hash_dic[id.ID], tuple):
                            if self._print:
                                print((
                                    "Collision! You are trying to fill {} again with A1_{}/A2_{},"
                                    " it is already inserted with A1_{}/A2_{}. "
                                    "You can resolve the collisions, use genotype data with CHR and bp info".format(
                                        id.ID,
                                        id['allele1'], id['allele2'],
                                        self.hash_dic[id.ID][1], self.hash_dic[id.ID][2])

                                ))
                            self.hash_dic[id.ID] = [self.hash_dic[id.ID], (self.n_keys, id['allele1'], id['allele2'])]
                            self.n_keys += 1
                        else:
                            self.hash_dic[id.ID].append((self.n_keys, id['allele1'], id['allele2']))
                            self.n_keys += 1
            except:

                if self.hash_dic.get(id.ID) is None:
                    self.hash_dic[id.ID] = [self.n_keys, id['allele1'], id['allele2']]
                    self.n_keys += 1
                else:
                    if isinstance(self.hash_dic[id[1].ID], tuple):
                        if self._print:
                            print((
                                "Collision! You are trying to fill {} again with A1_{}/A2_{},"
                                " it is already inserted with A1_{}/A2_{}. "
                                "You can resolve the collisions, use genotype data with CHR and bp info".format(
                                    id.ID,
                                    id['allele1'], id['allele2'],
                                    self.hash_dic[id.ID][1], self.hash_dic[id.ID][2])

                            ))
                        self.hash_dic[id.ID] = [self.hash_dic[id.ID], (self.n_keys, id['allele1'], id['allele2'])]
                        self.n_keys += 1
                    else:
                        self.hash_dic[id.ID].append((self.n_keys, id['allele1'], id['allele2']))
                        self.n_keys += 1

            # self.protected=True
        else:
            raise ValueError('Hash table already filled!')

    def fill(self, df):
        if not self.protected:
            for i in df.iterrows():
                try:
                    if i[1].ID[:2] == 'rs':
                        m = int(i[1].ID[2:])
                        if self.hash_table[m] != 0:
                            print('collision {}!'.format(i))
                            if isinstance(self.list[self.hash_table[m]], tuple):
                                self.list[self.hash_table[m]] = [self.list[self.hash_table[m]],
                                                                 (i[0], i[1]['allele1'], i[1]['allele2'])]
                            else:
                                self.list[self.hash_table[m]].append((i[0], i[1]['allele1'], i[1]['allele2']))
                        else:
                            self.list.append((i[0], i[1]['allele1'], i[1]['allele2']))
                            self.hash_table[m] = len(self.list)
                    else:
                        if self.hash_dic.get(i[1].ID) is None:
                            self.hash_dic[i[1].ID] = (i[0], i[1]['allele1'], i[1]['allele2'])
                        else:
                            if isinstance(self.hash_dic[i], tuple):
                                print((
                                    "Collision! You are trying to fill {} again with A1_{}/A2_{},"
                                    " it is already inserted with A1_{}/A2_{}. "
                                    "You can resolve the collisions, use genotype data with CHR and bp info".format(
                                        i[1].ID,
                                        i[1]['allele1'], i['allele2'],
                                        self.hash_dic[i[1].ID][1], self.hash_dic[i[1].ID][2])

                                ))
                                self.hash_dic[i[1].ID] = [self.hash_dic[i[1].ID],
                                                          (i[0], i[1]['allele1'], i[1]['allele2'])]
                            else:
                                self.hash_dic[i[1].ID].append((i[0], i[1]['allele1'], i[1]['allele2']))
                except:
                    if self.hash_dic.get(i[1].ID) is None:
                        self.hash_dic[i[1].ID] = [i[0], i[1]['allele1'], i[1]['allele2']]
                    else:
                        if isinstance(self.hash_dic[i[1].ID], tuple):
                            print((
                                "Collision! You are trying to fill {} again with A1_{}/A2_{},"
                                " it is already inserted with A1_{}/A2_{}. "
                                "You can resolve the collisions, use genotype data with CHR and bp info".format(
                                    i[1].ID,
                                    i[1]['allele1'], i[1]['allele2'],
                                    self.hash_dic[i[1].ID][1], self.hash_dic[i[1].ID][2])

                            ))
                            self.hash_dic[i[1].ID] = [self.hash_dic[i[1].ID], (i[0], i[1]['allele1'], i[1]['allele2'])]
                        else:
                            self.hash_dic[i[1].ID].append((i[0], i[1]['allele1'], i[1]['allele2']))

            self.protected = True
        else:
            raise ValueError('Hash table already filled!')

    def get_index(self, value):
        try:
            if value[:2] == 'rs':
                v = int(value[2:])
                r = int(self.hash_table[v])
                if r == 0:
                    return -1, 'R', 'R'
                else:
                    return self.list[r - 1]
            else:
                return self.hash_dic[value]
        except:
            return -1, 'R', 'R'

    def get_map(self, info):

        def _map(r, a1, a2):
            ind, A1, A2 = r
            if ind == -1:
                result = [-1, 0]
            elif A1 == a1 and A2 == a2:
                result = [ind, 1]
            elif A1 == a2 and A2 == a1:
                result = [ind, -1]
            else:
                result = [-1, 0]
            return result

        r = self.get_index(info.ID)
        if not isinstance(r[0], tuple):
            return _map(r, info.allele1, info.allele2)
        else:
            for i in r:
                m = _map(i, info.allele1, info.allele2)
                if m[0] != -1:
                    return m
            return [-1, 0]

    def save(self, path, filename):
        with open(os.path.join(path, filename), 'wb') as output:
            pickle.dump(self, output, -1)


if __name__ == "__main__":
    print('Hash')



import os
import sys
import warnings
import shutil
import matplotlib
import datetime
warnings.filterwarnings('ignore')
matplotlib.use('agg')
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import tensorflow as tf
import tensorflow.keras as K

tf.keras.backend.set_epsilon(0.0000001)
from gennet.utils.Dataloader import *
from gennet.utils.Utility_functions import *
from gennet.utils.Create_network import *
from gennet.utils.Create_plots import *
from gennet.utils.Utility_functions import load_train_arguments


def weighted_binary_crossentropy(y_true, y_pred):
    y_true = K.backend.clip(tf.cast(y_true, dtype=tf.float32), 0.0001, 1)
    y_pred = K.backend.clip(tf.cast(y_pred, dtype=tf.float32), 0.0001, 1)

    return K.backend.mean(
        -y_true * K.backend.log(y_pred + 0.0001) * weight_positive_class - (1 - y_true) * K.backend.log(
            1 - y_pred + 0.0001) * weight_negative_class)


def train_model(args):
    args.SlURM_JOB_ID = get_SLURM_id()
    model = None
    masks = None

    
    args.datapath = args.path
    

    if args.genotype_path == "undefined":
        args.genotype_path = args.path

    if args.mixed_precision:
        use_mixed_precision()

    args.multiprocessing = True if args.workers > 1 else False

    check_data(datapath=args.path, genotype_path=args.genotype_path, mode=args.problem_type)

    global weight_positive_class, weight_negative_class

    weight_positive_class = args.wpc
    weight_negative_class = 1

    args.train_size = sum(pd.read_csv(args.path + "subjects.csv")["set"] == 1)
    args.val_size = sum(pd.read_csv(args.path + "subjects.csv")["set"] == 2)
    args.test_size = sum(pd.read_csv(args.path + "subjects.csv")["set"] == 3)
    args.num_covariates = pd.read_csv(args.path + "subjects.csv").filter(like='cov_').shape[1]
    args.val_size_train = args.val_size

    if args.epoch_size is None:
        args.epoch_size = args.train_size
    else:
        args.val_size_train = min(args.epoch_size // 2, args.val_size)
        print("Using each epoch", args.epoch_size, "randomly selected training examples")
        print("Validation set size used during training is also set to half the epoch_size")

    args.inputsize = get_inputsize(args.genotype_path)

    folder, args.resultpath = get_paths(args)

    print("weight_positive_class", weight_positive_class)
    print("weight_negative_class", weight_negative_class)
    print("jobid =  " + str(args.ID))
    print("folder = " + str(folder))
    print("batchsize = " + str(args.batch_size))
    print("lr = " + str(args.learning_rate))

    model, masks = get_network(args)


    csv_logger = K.callbacks.CSVLogger(args.resultpath + 'train_log.csv', append=True)

    early_stop = K.callbacks.EarlyStopping(monitor='val_loss', min_delta=0.0001, patience=args.patience, verbose=1,
                                           mode='auto', restore_best_weights=True)
    save_best_model = K.callbacks.ModelCheckpoint(args.resultpath + "bestweights_job.h5", monitor='val_loss',
                                                  verbose=1, save_best_only=True, mode='auto')

    reduce_lr = K.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=0.2,
                                              patience=5, min_lr=0.001)

    if os.path.exists(args.resultpath + '/bestweights_job.h5') and not(args.resume):
        print('Model already Trained')
    elif os.path.exists(args.resultpath + '/bestweights_job.h5') and args.resume:
        print("load and save weights before resuming")
        shutil.copyfile(args.resultpath + '/bestweights_job.h5', args.resultpath + '/weights_before_resuming_' 
                        + datetime.datetime.now().strftime("%Y_%m_%d-%I_%M_%p")+'.h5') # save old weights
        log_file = pd.read_csv(args.resultpath + "/train_log.csv")
        save_best_model = K.callbacks.ModelCheckpoint(args.resultpath + "bestweights_job.h5", monitor='val_loss',
                                                  verbose=1, save_best_only=True, mode='auto', 
                                                      initial_epoch=len(log_file))
            
        print("Resuming training")

        model.load_weights(args.resultpath + '/bestweights_job.h5')
        train_generator = TrainDataGenerator(datapath=args.path,
                                             genotype_path=args.genotype_path,
                                             batch_size=args.batch_size,
                                             trainsize=int(args.train_size),
                                             inputsize=args.inputsize,
                                             epoch_size=args.epoch_size,
                                             one_hot=args.onehot)

        history = model.fit_generator(
            generator=train_generator,
            shuffle=True,
            epochs=args.epochs,
            verbose=1,
            callbacks=[early_stop, save_best_model, csv_logger, reduce_lr],
            workers=args.workers,
            use_multiprocessing=args.multiprocessing,
            validation_data=EvalGenerator(datapath=args.path, genotype_path=args.genotype_path, batch_size=args.batch_size,
                                          setsize=args.val_size_train,
                                          inputsize=args.inputsize, evalset="validation")
        )
    else:
        print("Start training from scratch")
        train_generator = TrainDataGenerator(datapath=args.path,
                                             genotype_path=args.genotype_path,
                                             batch_size=args.batch_size,
                                             trainsize=int(args.train_size),
                                             inputsize=args.inputsize,
                                             epoch_size=args.epoch_size,
                                             one_hot=args.onehot)

        history = model.fit_generator(
            generator=train_generator,
            shuffle=True,
            epochs=args.epochs,
            verbose=1,
            callbacks=[early_stop, save_best_model, csv_logger, reduce_lr],
            workers=args.workers,
            use_multiprocessing=args.multiprocessing,
            validation_data=EvalGenerator(datapath=args.path, genotype_path=args.genotype_path, batch_size=args.batch_size,
                                          setsize=args.val_size_train, one_hot=args.onehot,
                                          inputsize=args.inputsize, evalset="validation")
        )

    plot_loss_function(args.resultpath)
    model.load_weights(args.resultpath + '/bestweights_job.h5')
    print("Finished")
    
    save_train_arguments(args)


    if args.regression:
        # Regression-specific post training analysis
        print("Analysis over the validation set")
        pval = model.predict_generator(
            EvalGenerator(datapath=args.datapath, genotype_path=args.genotype_path, batch_size=args.batch_size,
                          setsize=args.val_size, inputsize=args.inputsize, evalset="validation", one_hot=args.onehot))
        yval = get_labels(args.datapath, set_number=2)
        fig_val, mse_val, explained_variance_val, r2_val = evaluate_performance_regression(yval, pval)
        np.save(args.resultpath + "/pval.npy", pval)
        fig_val.savefig(args.resultpath + "/validation_predictions.png", bbox_inches='tight', pad_inches=0)

        print("Analysis over the test set")
        ptest = model.predict_generator(
            EvalGenerator(datapath=args.datapath, genotype_path=args.genotype_path, batch_size=args.batch_size,
                          setsize=args.test_size, inputsize=args.inputsize, evalset="test", one_hot=args.onehot))
        ytest = get_labels(args.datapath, set_number=3)
        fig_test, mse_test, explained_variance_test, r2_test = evaluate_performance_regression(ytest, ptest)
        np.save(args.resultpath + "/ptest.npy", ptest)
        fig_test.savefig(args.resultpath + "/test_predictions.png", bbox_inches='tight', pad_inches=0)
    else:
        # Classification-specific post training analysis
        print("Analysis over the validation set")
        pval = model.predict_generator(
            EvalGenerator(datapath=args.datapath, genotype_path=args.genotype_path, batch_size=args.batch_size,
                          setsize=args.val_size, evalset="validation", inputsize=args.inputsize, one_hot=args.onehot))
        yval = get_labels(args.datapath, set_number=2)
        auc_val, confusionmatrix_val = evaluate_performance_classification(yval, pval)
        

        print("Analysis over the test set")
        ptest = model.predict_generator(
            EvalGenerator(datapath=args.datapath, genotype_path=args.genotype_path, batch_size=args.batch_size,
                          setsize=args.test_size, evalset="test", inputsize=args.inputsize, one_hot=args.onehot))
        ytest = get_labels(args.datapath, set_number=3)
        auc_test, confusionmatrix_test = evaluate_performance_classification(ytest, ptest)
        

    # Saving Results
    data = {'Jobid': args.ID,
            'Datapath': str(args.path),
            'genotype_path': str(args.genotype_path),
            'Batchsize': args.batch_size,
            'Learning rate': args.learning_rate,
            'L1 value': args.L1,
            'L1 act': args.L1_act,
            'patience': args.patience,
            'epoch size': args.epoch_size,
            'epochs': args.epochs,
            'onehot': args.onehot,
            'SlURM_JOB_ID': args.SlURM_JOB_ID}
    
    
    if args.regression:
        data.update({
            'MSE validation': mse_val,
            'MSE test': mse_test,
            'Explained variance val': explained_variance_val,
            'Explained variance test': explained_variance_test,
            'R2_validation': r2_val,
            'R2_test': r2_test
        })
    else:
        data.update({
            'Weight positive class': args.wpc,
            'AUC validation': auc_val,
            'AUC test': auc_test,
        })

        data['confusionmatrix_val'] = confusionmatrix_val
        data['confusionmatrix_test'] = confusionmatrix_test

    pd_summary_row = pd.Series(data)
    pd_summary_row.to_csv(args.resultpath + "/pd_summary_results.csv")
    
    with open(args.resultpath + "results_summary.txt", 'w') as f: 
        for key, value in data.items(): 
            f.write('%s:%s\n' % (key, value))

    if os.path.exists(args.datapath + "/topology.csv"):
        importance_csv = create_importance_csv(args.datapath, model, masks)
        importance_csv.to_csv(args.resultpath + "connection_weights.csv")

def get_network(args):
    """needs the following inputs
        args.inputsize
        args.L1
        args.L1_act
        args.datapath
        args.genotype_path
        args.num_covariates
        args.filter
    """
    regression = args.regression if hasattr(args, 'regression') else False
    args.init_linear = args.init_linear if hasattr(args, 'init_linear') else False
    args.L1 = args.L1 if hasattr(args, 'L1') else 0
    args.L1_act = args.L1_act if hasattr(args, 'L1_act') else 0

    global weight_positive_class, weight_negative_class

    weight_positive_class = args.wpc if hasattr(args, 'wpc') else 1
    weight_negative_class = 1

    if args.network_name == "lasso" and not regression:
        print("lasso network")
        model, masks = lasso(inputsize=args.inputsize, l1_value=args.L1, L1_act=args.L1_act if hasattr(args, 'L1_act') else None)

    elif args.network_name == "sparse_directed_gene_l1" and not regression:
        print("sparse_directed_gene_l1 network")
        model, masks = sparse_directed_gene_l1(inputsize=args.inputsize, l1_value=args.L1, one_hot=args.onehot)

    elif args.network_name == "regression_height" and regression:
        print("regression_height network")
        model, masks = regression_height(inputsize=args.inputsize, l1_value=args.L1)

    elif args.network_name == "gene_network_multiple_filters":
        print("gene_network_multiple_filters network")
        model, masks = gene_network_multiple_filters(datapath=args.datapath, inputsize=args.inputsize, genotype_path=args.genotype_path,
                                                     l1_value=args.L1, L1_act=args.L1_act, 
                                                     regression=regression, num_covariates=args.num_covariates,
                                                     filters=args.filters, one_hot=args.onehot)

    elif args.network_name == "gene_network_snp_gene_filters":
        print("gene_network_snp_gene_filters network")
        model, masks = gene_network_snp_gene_filters(datapath=args.datapath, inputsize=args.inputsize, genotype_path=args.genotype_path,
                                                     l1_value=args.L1, L1_act=args.L1_act, 
                                                     regression=regression, num_covariates=args.num_covariates,
                                                     filters=args.filters, one_hot=args.onehot)
    else:
        if os.path.exists(args.datapath + "/topology.csv"):
            model, masks = create_network_from_csv(datapath=args.datapath, inputsize=args.inputsize, genotype_path=args.genotype_path,
                                                   l1_value=args.L1, L1_act=args.L1_act, regression=regression,
                                                   num_covariates=args.num_covariates, one_hot=args.onehot)
        elif len(glob.glob(args.datapath + "/*.npz")) > 0:
            model, masks = create_network_from_npz(datapath=args.datapath, inputsize=args.inputsize, genotype_path=args.genotype_path,
                                                   l1_value=args.L1, L1_act=args.L1_act, regression=regression,
                                                   num_covariates=args.num_covariates, one_hot=args.onehot,
                                                   mask_order=args.mask_order if hasattr(args, 'mask_order') else None)

    optimizer_model = tf.keras.optimizers.Adam(lr=args.learning_rate)

    if regression:
        model.compile(loss="mse", optimizer=optimizer_model, metrics=["mse"])
    else:
        model.compile(loss=weighted_binary_crossentropy, optimizer=optimizer_model,
                      metrics=["accuracy", sensitivity, specificity])

    with open(args.resultpath + '/model_architecture.txt', 'w') as fh:
        model.summary(print_fn=lambda x: fh.write(x + '\n'))


    if args.init_linear and not(args.onehot):
        print("init_linear is only used in combination with the onehot flag")
        exit()
    if args.onehot and args.init_linear:        
        local = model.layers[1].get_weights()[0]
        
        if len(local.shape)  == 3:
            local_base = np.array(local[:,0,0], copy=True) 
            for i in range(local.shape[1]):
                local[:,i,0] = local_base + i 

        elif len(local.shape) == 1:
            local_base = np.array(local, copy = True)
            N = local_base.shape[0] // 3
            local_add = np.repeat([0, 1, 2], N)
            local = local_base + local_add
        else:
            print("length of local seems off")
            
        model.layers[1].set_weights([local, model.layers[1].get_weights()[1]])    


    return model, masks


def load_trained_network(args):
    """
    args.resultpath

    get_network needs the following inputs
        args.inputsize
        args.L1
        args.L1_act
        args.datapath
        args.genotype_path
        args.num_covariates
        args.filter
    """

    args = load_train_arguments(args)

    model, mask = get_network(args)
    model.load_weights(args.resultpath + '/bestweights_job.h5')

    return model, mask







import os
import sys

# os.chdir('../GenNet/')

# print(os.getcwd())
# sys.path.insert(1, os.getcwd())
# # sys.path.insert(1, os.getcwd() + "/GenNet_utils/")

import pytest
import pandas as pd
import shutil
from os.path import dirname, abspath
import argparse

print(sys.path)
from gennet.utils.Create_plots import plot
# from GenNet_utils.Train_network import train_model
from gennet.utils.Convert import convert
from gennet.utils.Topology import topology


class ArgparseSimulatorConvert():
    def __init__(self,
                 mode='/',
                 genotype=['examples/plink2/'],
                 study_name=['toy_data'],
                 outfolder="processed_data/",
                 step = "all"):
        self.mode = mode
        self.genotype = genotype
        self.study_name = study_name
        self.out = outfolder
        self.step = step
        self.vcf = False
        self.variants = None
        self.tcm = 500000000
        self.n_jobs = 1
        self.comp_level = 1


args = ArgparseSimulatorConvert()

convert(args)

        



import statsmodels.api as sm
import tqdm
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import tensorflow as tf
import os
import tables
import tensorflow.keras as K
import sklearn.metrics as skm
import scipy
import sys
import argparse
import time
import tqdm
import glob
import scipy
import scipy.sparse

from collections import defaultdict


def calculate_RLIPP_complete(model, xtrain, ytrain,input_regression_path, child_layer=None, masks=None, verbose = False):
 
    '''returns 
    
    1) new RLIPP: pd_dataframe with columns: [parent_layer_name	parent_layer	child_node	adjR_squar_parent	adjR_squar_child	RLIPP	parent_node	abs(RLIPP)]
    2) old RLIPP: pd_dataframe with columns: ['layer_name', 'layer_num', 'parent_node', 'child_node' 'RLIPP']
    3) child_dic: a dictionary with all the connections between parents and children 
    
    mode: tf model
    child_layer: layer to calculate RLIPP for
    masks: list of sparse coo_matrix masks
    '''
    pd_data_rlipp_old, child_dic =  calculate_RLIPP(model, xtrain, ytrain,input_regression_path, masks=masks)
    pd_data_rlipp_novel =  calculate_RLIPP_novel(model, xtrain, ytrain,input_regression_path,child_dic,  masks=masks)
    return pd_data_rlipp_novel, pd_data_rlipp_old, child_dic# , pd_novel

    
def calculate_RLIPP(model, xtrain, ytrain,input_regression_path, child_layer=None, masks=None ):
    
    '''returns pd_dataframe with columns: ['layer_name', 'layer_num', 'parent_node', 'child_node' 'RLIPP']
    
    mode: tf model
    child_layer: layer to calculate RLIPP for
    masks: list of sparse coo_matrix masks
    '''

    child_dict = {}
    initial = None
    RLIPP_func = None
    pd_current = pd.DataFrame([], columns= ['parent_layer_name','parent_layer', 'child_node', 'RLIPP'])
    
    
    if child_layer is None: 
        layer_names = [x.name for x in model.layers if "activation" in x.name]
        layer_names
        assert len([x.__class__.__name__ for x in model.layers if "LocallyDirected" in x.__class__.__name__]) == len(masks)


    model  = get_node_correlation(model, xtrain, ytrain, layer_names=None)
    model   = get_input_correlation(model, xtrain, ytrain, input_regression_path) # only once per simulation
  

    for layer_num, layer in enumerate(model.layers):

        layer_type = model.layers[layer_num].__class__.__name__ 

        if ("activation" in layer.name) and ("snp" not in layer.name):
            pd_current, child_dict = RLIPP_func(model, layer_num, pd_current, masks, child_dict)
        elif "LocallyDirected1D" in layer_type:  
            print("Directed1D")
            RLIPP_func = get_RLIPP_LD
        elif "LocallyConnected1D" in layer_type :
            print("Connected1D")
            RLIPP_func = get_RLIPP_LC
        elif "Dense" in layer_type :
            print("Dense")
            RLIPP_func = get_RLIPP_dense
        else:
            print(" -- Skipping", layer_type, '--') 
          
    return pd_current, child_dict# , pd_novel
        

def calculate_RLIPP_novel(model, xtrain, ytrain,input_regression_path,child_dic, child_layer=None, masks=None, verbose = False ):
    model  = get_node_correlation_novel(model, xtrain, ytrain,child_dic,verbose, layer_names=None)

    pd_novel = pd.DataFrame([], columns= ['parent_layer_name','parent_layer', 'child_node','adjR_squar_parent', 'adjR_squar_child', 'RLIPP'])
    for layer_num, layer in enumerate(model.layers):
        layer_type = model.layers[layer_num].__class__.__name__
        if ("activation" in layer.name) and ("snp" not in layer.name):
            for node_num, el in enumerate(model.layers[layer_num].fprsq):
                child_prsqrt = model.layers[layer_num].fprsq[node_num]
                parent_prsqrt = model.layers[layer_num].prsqrt[node_num]
                RLIPP_score = float((parent_prsqrt - child_prsqrt) / child_prsqrt)
                if parent_prsqrt < 0:
                    RLIPP_score = 0.
                node_num_parent = node_num # -10 ## not removed the 10
                if verbose:
                    print("node_num ",  node_num_parent)
                pd_novel = pd_novel.append({"parent_layer":layer_num, 
                                    "parent_layer_name":model.layers[layer_num].name,
                                    "child_node":"NA",
                                    "parent_node":node_num_parent,
                                    "adjR_squar_parent": parent_prsqrt, 
                                    "adjR_squar_child": child_prsqrt,         
                                    "RLIPP":RLIPP_score}, ignore_index=True)
    return pd_novel


def get_input_correlation(model, xtrain, ytrain, input_regression_path):
    
    
    if os.path.exists(input_regression_path + '/input_correlation.npy') & False:
        model.layers[0].prsqrt = np.load(input_regression_path + '/input_correlation.npy')
        return model
  
    print("calculate input correlation")
    assert xtrain.ndim == 3, "xtrain needs to be 3 dimensional with (npat, nfeatures, 3 or 1)"
    
    prsqrt_input = np.full((xtrain.shape[1], xtrain.shape[2]), np.NaN)


    for SNP_num in tqdm.tqdm(range(xtrain.shape[1])):
        for allele in range(xtrain.shape[2]):
            try:
                regresult = sm.OLS(ytrain, xtrain[:,SNP_num, allele]).fit(disp=0)
                prsqrt_input[SNP_num, allele] =regresult.rsquared_adj
            except:
                prsqrt_input[SNP_num, allele] = np.NaN
    
    model.layers[0].prsqrt = prsqrt_input
    np.save(input_regression_path + '/input_correlation.npy', prsqrt_input)
                
    return model


def get_node_correlation(model, xtrain, ytrain, layer_names=None):
    for layer_num, layer in enumerate(model.layers):
        if 'activation' in layer.name:
            intermediate_layer_model = tf.keras.Model(inputs=model.input,
                                             outputs=model.get_layer(layer.name).output)
            predictions = intermediate_layer_model.predict(np.squeeze(xtrain))
            predictions = np.reshape(predictions, (xtrain.shape[0], -1))
            prsqrt_node = np.full((predictions.shape[1]), np.NaN)
            
            for node in range(predictions.shape[1]):
                try:
                    if len(np.unique(predictions[:,node])) > 1:
                        regresult = sm.OLS(ytrain, predictions[:,node]).fit(disp=0)
                        prsqrt_node[node] = regresult.rsquared_adj
                    else:
                        prsqrt_node[node] = np.NaN
                except:
                    prsqrt_node[node] = np.NaN
         
        else:
            prsqrt_node = np.full((0), np.NaN)
            
        model.layers[layer_num].prsqrt = prsqrt_node
        
    return model

def get_node_correlation_novel(model, xtrain, ytrain, child_dic, verbose, layer_names=None):
    for layer_num, layer in enumerate(model.layers):
        if 'activation' in layer.name:
                intermediate_layer_model = tf.keras.Model(inputs=model.input,
                                                 outputs=model.get_layer(layer.name).output)
                print(layer_num)
                try: 
                    previous_predictions = predictions
                except: 
                    if verbose:
                        print("1st")
                predictions = intermediate_layer_model.predict(np.squeeze(xtrain))
                pre_resh_shap = predictions.shape
                if verbose:
                    print("prediction = ", predictions.shape)
                predictions = np.reshape(predictions, (xtrain.shape[0], -1))


                prsqrt_node = np.full((predictions.shape[1]), np.NaN)
                prsqrt_node = np.full((predictions.shape[1]), np.NaN)
                fprsq_node  = np.full((predictions.shape[1]), np.NaN)
                for node in range(predictions.shape[1]):
                            if layer_num > 2: 
                                previous_predictions_ = np.reshape(previous_predictions, (xtrain.shape[0], -1))#
                                if verbose:
                                    print(previous_predictions_.shape) # (1000, 100)
                                    print(fprsq_node.shape)
                            try:
                                if len(np.unique(predictions[:,node])) > 1:
                                    regresult = sm.OLS(ytrain, predictions[:,node]).fit(disp=0)
                                    prsqrt_node[node] = regresult.rsquared_adj
                                    if layer_num > 2:  
                                        coef_flatter = (node // pre_resh_shap[1] )*pre_resh_shap[1]
                                        if verbose:
                                            print("coef_flatter", coef_flatter)                                    
                                        regresult_child = sm.OLS(ytrain, previous_predictions_[:, child_dic[layer_num][node-coef_flatter]]).fit(disp=0)
                                        #print(regresult_child.rsquared_adj)
                                        fprsq_node[node] = regresult_child.rsquared_adj
                                        if verbose:    
                                            print(fprsq_node)
                                       

                                    if verbose:
                                        print("layer_num", layer_num)
                                else:
                                    if verbose:
                                        print("else")
                                    prsqrt_node[node] = np.NaN
                                    fprsq_node[node] = np.NaN
                            except:
                                print("excep")
                                prsqrt_node[node] = np.NaN
                                fprsq_node[node] = np.NaN

        else:
            prsqrt_node = np.full((0), np.NaN)
            fprsq_node = np.full((0), np.NaN)

        model.layers[layer_num].prsqrt = prsqrt_node
        if layer_num > 2: 
            if verbose:
                print(fprsq_node)
            model.layers[layer_num].fprsq = fprsq_node
    return model

def get_child_nodes(model, layer_num):
    childeren_prsqrt = None
    
    for i in range(layer_num):
        if model.layers[i].prsqrt.size != 0:
            childeren_prsqrt = model.layers[i].prsqrt   
    return childeren_prsqrt


def get_RLIPP(model, pd_RLIPP, parent_prsqrt, child_prsqrt, node_num_child, node_num_parent, layer_num):
    RLIPP_score = float((parent_prsqrt - child_prsqrt) / child_prsqrt)
    if parent_prsqrt < 0:
        RLIPP_score = 0.
    pd_RLIPP = pd_RLIPP.append({"parent_layer":layer_num, 
                            "parent_layer_name":model.layers[layer_num].name,
                            "child_node":node_num_child,
                            "parent_node":node_num_parent,
                            "RLIPP":RLIPP_score}, ignore_index=True)
    return pd_RLIPP



def get_RLIPP_LC(model, layer_num, pd_RLIPP, masks, child_dict):
    parent_prsqrt = model.layers[layer_num].prsqrt
    childeren_prsqrt = get_child_nodes(model, layer_num)
    child_index = np.arange(len(childeren_prsqrt))
    
    kernelsize = int(len(childeren_prsqrt) / len(parent_prsqrt))
    
    childeren_prsqrt = np.reshape(childeren_prsqrt, (-1, kernelsize))
    
    child_dict[layer_num] = {}  
    child_index = np.reshape(child_index, (-1, kernelsize))
    

    # calculate the RLIPP for the right parent-child
    nodenum_child = 0
    for node_num, child_prsqrts in enumerate(childeren_prsqrt):
        child_dict[layer_num][node_num] = child_index[node_num]
        for child_prsqrt in child_prsqrts:
            
            pd_RLIPP = get_RLIPP(model, pd_RLIPP, parent_prsqrt[node_num], child_prsqrt, 
                                 node_num_child = nodenum_child, node_num_parent = node_num, layer_num = layer_num)  
            nodenum_child = nodenum_child + 1
        
    
    return pd_RLIPP, child_dict


def get_RLIPP_LD(model, layer_num, pd_RLIPP, masks, child_dict):
    parent_prsqrt = model.layers[layer_num].prsqrt
    childeren_prsqrt = get_child_nodes(model, layer_num)
    
    child_dict[layer_num] = {}                                          # to change


    num_mask = -1
    for i in range(layer_num):
        if "LocallyDirected" in model.layers[i].__class__.__name__:
            num_mask +=1

    mask = masks[num_mask]

    nfilters = int(len(parent_prsqrt) / mask.shape[1])
    parent_prsqrt = np.reshape(parent_prsqrt, (mask.shape[1],nfilters))

    nodenum_child = 0
    for cfilter in range(nfilters):                       # per filter
        cparent_prsqrt =  parent_prsqrt[:,cfilter]        # select the parent node
        for node_num in range(mask.shape[1]):            # per gene
            children = mask.row[mask.col==node_num]      # select children
            child_dict[layer_num][node_num] = children
            for child_prsqrt in childeren_prsqrt[children]:  # per child
                pd_RLIPP = get_RLIPP(model, pd_RLIPP, cparent_prsqrt[node_num], child_prsqrt, 
                                     node_num_child = nodenum_child, node_num_parent = node_num, layer_num = layer_num)
                nodenum_child = nodenum_child + 1

    return pd_RLIPP, child_dict



def get_RLIPP_dense(model, layer_num, pd_RLIPP, masks, child_dict):
    parent_prsqrt = model.layers[layer_num].prsqrt
    childeren_prsqrt = get_child_nodes(model, layer_num)
    
    child_dict[layer_num] = {}
    child_dict[layer_num][0] = list(np.arange(len(childeren_prsqrt)))
    
    # calculate the RLIPP for the right parent-child
    for node_num, child_prsqrt in enumerate(childeren_prsqrt):
        pd_RLIPP = get_RLIPP(model, pd_RLIPP, parent_prsqrt, child_prsqrt, 
                             node_num_child = node_num, node_num_parent= 0, layer_num=layer_num)

    return pd_RLIPP, child_dict
    
    
    

def get_RLIPP_dense(model, layer_num, pd_RLIPP, masks, child_dict):
    parent_prsqrt = model.layers[layer_num].prsqrt
    childeren_prsqrt = get_child_nodes(model, layer_num)
    
    child_dict[layer_num] = {}
    child_dict[layer_num][0] = list(np.arange(len(childeren_prsqrt)))
    
    # calculate the RLIPP for the right parent-child
    for node_num, child_prsqrt in enumerate(childeren_prsqrt):
        pd_RLIPP = get_RLIPP(model, pd_RLIPP, parent_prsqrt, child_prsqrt, 
                             node_num_child = node_num, node_num_parent= 0, layer_num=layer_num)

    return pd_RLIPP, child_dict






import seaborn as sns
import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import sklearn.metrics as skm
import scipy
import sys
import pandas as pd
import numpy as np
import itertools

from scipy.sparse import coo_matrix



def make_importance_values_input(model, masks):
    
    '''
    Regular importance for the input based on a multiplication over all the
    weights propogated to the input
    '''
    mask_count = 0
    previous_layer_type = ""
    importance = pd.DataFrame([])

    inputsize = model.layers[0].input_shape[0][1]
    importance["key_0"] = np.arange(inputsize)
    importance["importance_0"] = np.ones(inputsize)

    for i, layer in enumerate(model.layers):
        importance_new = pd.DataFrame([])
        layer_type = model.layers[i].__class__.__name__ 

        if "LocallyDirected1D" in layer_type:
            mask = masks[mask_count]
            mask_count +=1 

            weights = layer.get_weights()[0]
            names = ['weights_'+str(i)+ "_"+str(j) for j in range(weights.shape[1])]

            importance_layer = pd.DataFrame(weights, columns= names)
            importance_layer["key_0"] = mask.row
            importance_layer["key_1"] = mask.col


            importance = importance.merge(importance_layer, on="key_0")
            importance = importance.drop("key_0", axis=1)
            importance = importance.rename({"key_1":"key_0"}, axis=1)

            num_filter= 0
            for importance_col in importance.filter(like="importance_").columns:
                for weight_col  in importance_layer.filter(like="weights_").columns:
                    importance_new["importance_"+str(num_filter)] = importance[importance_col]*importance[weight_col]
                    num_filter+=1

            importance_new["key_0"] = importance_layer["key_1"]
            importance = importance_new
            previous_layer_type = layer_type
            print(layer_type, importance.shape)


        elif "LocallyConnected1D" in layer_type :

            importance_new = pd.DataFrame([])
            importance_temp = pd.DataFrame([])

            layer_type = model.layers[i].__class__.__name__ 

            if len(importance) > 1:
                pass
            else:
                print("Had expected a different layer before, not", previous_layer_type)
                print("asume 1x1")


            weights = np.reshape(layer.get_weights()[0], layer.input_shape[1:3] )
            names = ['weights_'+str(i)+ "_"+str(j) for j in range(weights.shape[1])]
            importance_layer = pd.DataFrame(weights, columns= names)
            importance_layer["key_0"] = np.arange(importance_layer.shape[0]) 

            importance = importance.merge(importance_layer, on="key_0")

            importance_new["importance_0"] = (importance.filter(like="importance").prod(axis=1)* 
                                              importance.filter(like="weights_").prod(axis=1))

            importance_new["key_0"] = importance["key_0"]
            importance = importance_new
            previous_layer_type = layer_type    
            print(layer_type, importance.shape)

        elif "Dense" in layer_type :

            importance_new = pd.DataFrame([])
            importance_temp = pd.DataFrame([])

            layer_type = model.layers[i].__class__.__name__ 

            if len(importance) > 1:
                pass
            else:
                print("Had expected a different layer before, not", previous_layer_type)
                print("asume 1x1")


            weights = np.reshape(layer.get_weights()[0], layer.input_shape[1:3] )
            names = ['weights_dense']
            importance_layer = pd.DataFrame(weights, columns= names)
            importance_layer["key_0"] = np.arange(importance_layer.shape[0]) 

            importance = importance.merge(importance_layer, on="key_0")

            importance_new["importance_0"] = (importance.filter(like="importance").prod(axis=1)* 
                                              importance.filter(like="weights_").prod(axis=1))

            importance_new["key_0"] = importance["key_0"]
            importance = importance_new
            previous_layer_type = layer_type    

            print(layer_type, importance.shape)        


        else:
            print(" -- Skipping", layer_type, '--')


        importance["percentage"] = abs(importance["importance_0"]) / sum(abs(importance["importance_0"]))
    return importance


import pandas as pd
import numpy as np
import glob
import itertools
import sys
import os
import time
import tqdm
import sklearn.metrics as skm


def get_auc_epistasis(true_label, predicted_label,  threshold=100, verbose=True):
       
    
    predicted_label = predicted_label.flatten()[:threshold]
    true_label = true_label.flatten()[:threshold]
    
   
    if (threshold > len(predicted_label) ) | (threshold > len(true_label) ):
        if verbose:
            print("WARNING: Threshold is too high, all labels are considered")
    
      
    if (len(np.unique(predicted_label)) < 2 ) | (len(np.unique(true_label)) < 2):
        if verbose:
            print("Error: all labels are the same")
        return 0
   
    fpr, tpr, thresholds = skm.roc_curve(true_label, predicted_label)
    roc_auc = skm.auc(fpr, tpr)
    
    return roc_auc


def get_GenNet_Gametes_epistasis_AUC(pd_overview, threshold=100, verbose=True):
    """pd_overview: pd dataframe with columns: rfrunpath, num_snps"""
    pd_interact = pd.DataFrame([])
    auc= np.zeros(pd_overview.shape[0])
    
    for i in tqdm.tqdm(range(len(pd_overview))):
        try:
            folder = pd_overview["rfrunpath"].iloc[i] +  "/NID.csv"
            pd_interact  = pd.read_csv(folder, names = ['interaction','strength'], skiprows=1)
            pd_interact[['X_1', 'X_2']] = pd_interact['interaction'].str.rstrip(' ').str.split(" ", expand=True)
            pd_interact["X_1"] = pd.to_numeric(pd_interact["X_1"])
            pd_interact["X_2"] = pd.to_numeric(pd_interact["X_2"])
            pd_interact["groundtruth"] = np.zeros(len(pd_interact), int)

            causal_snps = [pd_overview["num_snps"].iloc[i]-1, pd_overview["num_snps"].iloc[i]-2]
            pd_interact.loc[(pd_interact['X_1'].isin(causal_snps)) & (pd_interact['X_2'].isin(causal_snps)) &
                            (pd_interact['X_2'] != pd_interact['X_1']), "groundtruth"]=1

            pd_interact =  pd_interact.sort_values("strength", ascending=False)
            auc[i] = get_auc_epistasis(pd_interact["groundtruth"].values,pd_interact["strength"].values, 
                                       threshold=threshold, verbose=verbose)
        except:
            print(pd_overview["simu_name"].iloc[i])
            auc[i] = -1    

    pd_overview["NID eAUC"] = auc
    return pd_overview
    
    
def get_LGBM_Gametes_epistasis_AUC(pd_overview, threshold=100, verbose=True):
    """pd_overview: pd dataframe with columns: rfrunpath, num_snps"""
    pd_interact = pd.DataFrame([])
    auc= np.zeros(pd_overview.shape[0])
    
    for i in tqdm.tqdm(range(len(pd_overview))):
        try:
            pd_interact  = pd.read_csv("./Gametes_output_2D/"+pd_overview["simu_name"].iloc[i]+".csv", index_col = 0)
            pd_interact["groundtruth"] = np.zeros(len(pd_interact), int)
            causal_snps = ["M0P0","M0P1"]
            pd_interact.loc[(pd_interact['SNP1'].isin(causal_snps)) &
                            (pd_interact['SNP2'].isin(causal_snps)) & 
                            (pd_interact['SNP2'] != pd_interact['SNP1']), "groundtruth"]=1
            pd_interact =  pd_interact.sort_values("strength", ascending=False)
            auc[i] = get_auc_epistasis(pd_interact["groundtruth"].values, pd_interact["strength"].values, 
                                       threshold=threshold, verbose=verbose)

        except:
            print(pd_overview["simu_name"].iloc[i])
            auc[i] = -1    

    pd_overview["NID eAUC"] = auc
    return pd_overview
    




import os
import sys

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from utils.hase.config import PYTHON_PATH

if PYTHON_PATH is not None:
    for i in PYTHON_PATH: sys.path.insert(0, i)
import argparse
import h5py
import pandas as pd
import numpy as np
from utils.hase.hdgwas.tools import Timer
import tables


def probes_minimac2hdf5(data_path, save_path, study_name, chunk_size=1000000):
    if os.path.isfile(os.path.join(save_path, 'probes', study_name + '.h5')):
        os.remove(os.path.join(save_path, 'probes', study_name + '.h5'))

    hash_table = {'keys': np.array([], dtype=int), 'allele': np.array([])}

    df = pd.read_csv(data_path, sep=' ', chunksize=chunk_size, header=None, index_col=None)
    for i, chunk in enumerate(df):
        print('add chunk {}'.format(i))
        chunk.columns = ["ID", 'allele1', 'allele2', 'MAF', 'Rsq']
        hash_1 = chunk.allele1.apply(hash)
        hash_2 = chunk.allele2.apply(hash)
        k, indices = np.unique(np.append(hash_1, hash_2), return_index=True)
        s = np.append(chunk.allele1, chunk.allele2)[indices]
        ind = np.invert(np.in1d(k, hash_table['keys']))
        hash_table['keys'] = np.append(hash_table['keys'], k[ind])
        hash_table['allele'] = np.append(hash_table['allele'], s[ind])
        chunk.allele1 = hash_1
        chunk.allele2 = hash_2
        chunk.to_hdf(os.path.join(save_path, 'probes', study_name + '.h5'), data_columns=True, key='probes',
                     format='table', append=True,
                     min_itemsize=25, complib='zlib', complevel=9)
    pd.DataFrame.from_dict(hash_table).to_csv(os.path.join(save_path, 'probes', study_name + '_hash_table.csv.gz'),
                                              index=False, compression='gzip', sep='\t')


def ind_minimac2hdf5(data_path, save_path, study_name):
    if os.path.isfile(os.path.join(save_path, 'individuals', study_name + '.h5')):
        os.remove(os.path.join(save_path, 'individuals', study_name + '.h5'))
    n = []
    f = open(data_path, 'r')
    for i, j in enumerate(f):
        n.append((j[:-1]))
    f.close()
    n = np.array(n)
    chunk = pd.DataFrame.from_dict({"individual": n})
    chunk.to_hdf(os.path.join(save_path, 'individuals', study_name + '.h5'), key='individuals', format='table',
                 min_itemsize=25, complib='zlib', complevel=9)


def id_minimac2hdf5(data_path, id, save_path):
    n = []
    f = open(data_path, 'r')
    for i, j in enumerate(f):
        try:
            n.append(np.float(j))
        except:
            n.append(np.float(-1))
    n = np.array(n)
    f.close()
    store = h5py.File(os.path.join(save_path, 'genotype', id + '.h5'), 'w')
    with Timer() as t:
        store.create_dataset(id, data=n, compression='gzip', compression_opts=9)
    print('standard save gzip 9...', t.secs)
    store.close()


def id_minimac2hdf5_pandas(data_path, id, save_path):
    df = pd.read_csv(data_path, header=None, index_col=None)
    df.columns = ["genotype"]
    n = df["genotype"].as_matrix()
    store = h5py.File(os.path.join(save_path, 'genotype', id + '.h5'), 'w')
    with Timer() as t:
        store.create_dataset(id, data=n, compression='gzip', compression_opts=9)
    print('pandas save gzip 9...', t.secs)
    store.close()
    df = None


def genotype_minimac2hdf5(data_path, id, save_path, study_name):
    df = pd.read_csv(data_path, header=None, index_col=None, sep='\t', dtype=np.float16)
    data = df.as_matrix()
    data = data.T
    print(data.shape)
    print('Saving chunk...{}'.format(os.path.join(save_path, 'genotype', str(id) + '_' + study_name + '.h5')))
    h5_gen_file = tables.open_file(
        os.path.join(save_path, 'genotype', str(id) + '_' + study_name + '.h5'), 'w', title=study_name)

    atom = tables.Float16Atom()
    genotype = h5_gen_file.create_carray(h5_gen_file.root, 'genotype', atom,
                                         (data.shape),
                                         title='Genotype',
                                         filters=tables.Filters(complevel=9, complib='zlib'))
    genotype[:] = data
    h5_gen_file.close()
    os.remove(data_path)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Script to convert Minimac data')
    parser.add_argument("-study_name", required=True, type=str, help="Study specific name")
    parser.add_argument("-id", type=str, help="subject id")
    parser.add_argument("-data", required=True, type=str, help="path to file")
    parser.add_argument("-out", required=True, type=str, help="path to results save folder")
    parser.add_argument("-flag", required=True, type=str, choices=['genotype', 'individuals', 'probes', 'chunk'],
                        help="path to file with SNPs info")

    args = parser.parse_args()

    print(args)
    try:
        print('Creating directories...')
        os.makedirs(os.path.join(args.out, 'genotype'))
        os.makedirs(os.path.join(args.out, 'individuals'))
        os.makedirs(os.path.join(args.out, 'probes'))
        os.makedirs(os.path.join(args.out, 'tmp_files'))
    except:
        print(('Directories "genotype","probes","individuals" are already exist in {}...'.format(args.out)))

    if args.id is not None and args.flag == 'genotype':
        with Timer() as t:
            id_minimac2hdf5_pandas(args.data, args.id, args.out)
        print('time pandas...', t.secs)
    elif args.flag == 'probes':
        probes_minimac2hdf5(args.data, args.out, args.study_name)
    elif args.flag == 'individuals':
        ind_minimac2hdf5(args.data, args.out, args.study_name)
    elif args.flag == 'chunk':
        genotype_minimac2hdf5(args.data, args.id, args.out, args.study_name)



import gc
import os

import numpy as np
import pandas as pd
import tables


class Encoder(object):

    def __init__(self, out):
        self.name = 'Encoder'
        self.N_matrix = None
        self.F = None
        self.F_inv = None
        self.encoder_chunks = {}
        self.hdf5_iter = 0
        self.npy_iter = 0
        self.pytable_filters = tables.Filters(complevel=9, complib='zlib')
        self.out = out
        self.metadata = None
        self.study_name = None
        self.phen_info_dic = {}
        self.phen_info_dic['id'] = None
        try:
            print('Creating directories...')
            os.makedirs(os.path.join(self.out, 'encode_genotype'))
            os.makedirs(os.path.join(self.out, 'encode_phenotype'))
            os.makedirs(os.path.join(self.out, 'encode_individuals'))

        except:
            print((
                'Directories "encode_genotype","encode_phenotype","encode_individuals" are already exist in {}...'.format(
                    self.out)))

    def matrix(self, N, save=False):

        self.F = np.random.randint(1, 10, N * N).reshape(N, N)
        self.F_inv = np.linalg.inv(self.F)
        self.F = np.array(self.F, dtype=np.float)
        if save:
            print(('Saving decoders Matrices to...{}'.format(self.out)))
            np.save(os.path.join(self.out, 'F.npy'), self.F)
            np.save(os.path.join(self.out, 'F_inv.npy'), self.F_inv)

    def encode(self, data, data_type=None):

        if isinstance(self.F, type(None)):
            raise ValueError('Encode matrix is not define')

        x, y = data.shape
        # TODO (mid) add to encode_chunk info about data size and/or names indexes
        # TODO (mid) check the protocol if exist... all study on single stage

        a, b = self.F.shape

        if isinstance(data_type, type(None)):
            raise ValueError('data_type is None')
        elif data_type == "genotype":
            print('Decoding genotype')
            return np.dot(data, self.F)
        elif data_type == 'phenotype':
            print('Decoding phenotype')
            return np.dot(self.F_inv, data)

    def save_npy(self, data, save_path=None, info=None, index=None):
        # only for phenotype
        if isinstance(save_path, type(None)) or not os.path.isdir(save_path):
            raise ValueError('There is no such path or directory {}'.format(save_path))
        np.save(os.path.join(save_path, str(self.npy_iter) + '_' + self.study_name + '.npy'), data)
        if self.phen_info_dic['id'] is None:
            self.phen_info_dic['id'] = np.array(info._data.id)[index]
        self.phen_info_dic[str(self.npy_iter) + '_' + self.study_name + '.npy'] = info._data.names[
                                                                                  info._data.start:info._data.finish]
        self.npy_iter += 1

    def save_csv(self, data, save_path=None, info=None, index=None):
        if isinstance(save_path, type(None)) or not os.path.isdir(save_path):
            raise ValueError('There is no such path or directory {}'.format(save_path))
        df = pd.DataFrame(data)
        df.columns = info._data.names[info._data.start:info._data.finish]
        df.insert(0, 'id', np.array(info._data.id)[index])
        df.to_csv(os.path.join(save_path, str(self.npy_iter) + '_' + self.study_name + '.csv'), sep='\t', index=False)
        self.npy_iter += 1

    def save_hdf5(self, data, save_path=None, info=None, index=None):
        # only for genetics data
        print(('Saving data to ... {}'.format(
            os.path.join(save_path, str(self.hdf5_iter) + '_' + self.study_name + '_encoded.h5'))))

        if not os.path.isfile(os.path.join(self.out, 'encode_individuals', self.study_name + '.h5')):
            chunk = pd.DataFrame.from_dict({"individual": info._data.id[index]})
            chunk.to_hdf(os.path.join(self.out, 'encode_individuals', self.study_name + '.h5'), key='individuals',
                         format='table', min_itemsize=25, complib='zlib', complevel=9)

        if isinstance(save_path, type(None)) or not os.path.isdir(save_path):
            raise ValueError('There is no such path or directory {}'.format(save_path))

        self.h5_gen_file = tables.open_file(
            os.path.join(save_path, str(self.hdf5_iter) + '_' + self.study_name + '.h5'), 'w', title='encode_genotype')
        self.hdf5_iter += 1

        atom = tables.Float64Atom()
        self.genotype = self.h5_gen_file.create_carray(self.h5_gen_file.root, 'genotype', atom,
                                                       (data.shape),
                                                       title='Genotype', filters=self.pytable_filters)
        self.genotype[:] = data
        self.h5_gen_file.close()
        gc.collect()



import argparse
import os

import numpy as np
import pandas as pd

parser = argparse.ArgumentParser(description='')
parser.add_argument("-i", type=str, required=True, help="path to results file")
parser.add_argument("-p", type=str, required=True, help="path to probes file")
parser.add_argument("-o", type=str, required=True, help="path to output folder")
parser.add_argument("-split_ph", action='store_true', default=False,
                    help="flag to split result in separate files per phenotype")
parser.add_argument("-chunk", type=int, default=1000000, help="chunk size to read results file")
args = parser.parse_args()

store = pd.HDFStore(args.p, 'r')
df_probes_tmp = store.select('probes', start=0, stop=1)
probes_columns = df_probes_tmp.columns
if 'Rsq' in probes_columns:
    probes_columns = ['ID', 'Rsq', 'allele1', 'allele2']
else:
    probes_columns = ['ID', 'allele1', 'allele2']

df_probes = store.select('probes', columns=probes_columns)
print('Probes shape {}'.format(df_probes.shape))

if os.path.isfile(args.p.split('.h5')[0] + '_hash_table.csv.gz'):
    try:
        df_hash = pd.read_csv(args.p.split('.h5')[0] + '_hash_table.csv.gz', sep='\t',
                              compression='gzip', index_col=False)
    except:
        df_hash = pd.read_csv(args.p.split('.h5')[0] + '_hash_table.csv.gz', sep='\t',
                              index_col=False)

    print('Hash table found!')
    df_probes = pd.merge(df_probes, df_hash, right_on='keys', left_on='allele1')
    df_probes['allele1'] = df_probes['allele']
    del df_probes['allele']
    df_probes = pd.merge(df_probes, df_hash, right_on='keys', left_on='allele2')
    df_probes['allele2'] = df_probes['allele']
    df_probes = df_probes[probes_columns]
else:
    print('Hash table is not found!!!')

df = pd.read_csv(args.i, compression='infer', sep=" ", chunksize=args.chunk)

phenotypes = []
for i, df_chunk in enumerate(df):
    print('Chunk number {}'.format(i))
    df_chunk = pd.merge(df_chunk, df_probes, left_on='RSID', right_on='ID')
    df_chunk = df_chunk[probes_columns + ['MAF', 'BETA', 'SE', 'p_value', 'phenotype', 't-stat']]
    df_chunk.rename(columns={'MAF': 'AF_coded'}, inplace=True)
    if not args.split_ph:
        hdr = False if os.path.isfile(os.path.join(args.o, os.path.basename(args.i))) else True
        df_chunk.to_csv(os.path.join(args.o, os.path.basename(args.i)), sep=' ', index=None, mode='a', header=hdr)
    else:
        phenotypes = np.unique(np.append(phenotypes, df_chunk.phenotype))
        for ph in phenotypes:
            df_tmp = df_chunk.query('phenotype=={}'.format(ph))
            if df_tmp.shape[0] != 0:
                hdr = False if os.path.isfile(os.path.join(args.o, str(ph) + "_" + os.path.basename(args.i))) else True
                df_tmp[probes_columns + ['AF_coded', 'BETA', 'SE', 'p_value', 't-stat']].to_csv(
                    os.path.join(args.o, str(ph) + "_" + os.path.basename(args.i)), sep=' ', index=None, mode='a',
                    header=hdr)
